from __future__ import annotations

import logging
from abc import ABC
from datetime import date, datetime, timedelta, timezone
from math import floor

from soda_core.common.data_source_impl import DataSourceImpl
from soda_core.common.datetime_conversions import convert_str_to_datetime
from soda_core.common.logging_constants import soda_logger
from soda_core.common.sql_dialect import *
from soda_core.contracts.contract_verification import Check, CheckOutcome, CheckResult
from soda_core.contracts.impl.check_types.freshness_check_yaml import FreshnessCheckYaml
from soda_core.contracts.impl.check_types.row_count_check import RowCountMetricImpl
from soda_core.contracts.impl.contract_verification_impl import (
    AggregationMetricImpl,
    CheckImpl,
    CheckParser,
    ColumnImpl,
    ContractImpl,
    MeasurementValues,
    MetricImpl,
    ThresholdImpl,
)

logger: logging.Logger = soda_logger


class FreshnessCheckParser(CheckParser):
    def get_check_type_names(self) -> list[str]:
        return ["freshness"]

    def parse_check(
        self,
        contract_impl: ContractImpl,
        column_impl: Optional[ColumnImpl],
        check_yaml: FreshnessCheckYaml,
    ) -> Optional[CheckImpl]:
        return FreshnessCheckImpl(
            contract_impl=contract_impl,
            column_impl=column_impl,
            check_yaml=check_yaml,
        )


class FreshnessCheckImplBase(CheckImpl, ABC):
    def __init__(
        self,
        contract_impl: ContractImpl,
        column_impl: ColumnImpl,
        check_yaml: FreshnessCheckYaml,
    ):
        super().__init__(
            contract_impl=contract_impl,
            column_impl=column_impl,
            check_yaml=check_yaml,
        )

        self.now_variable: Optional[str] = check_yaml.now_variable
        self.unit: str = check_yaml.unit if check_yaml.unit else "hour"
        self.resolved_variable_values = contract_impl.contract_yaml.resolved_variable_values
        self.soda_variable_values = (
            contract_impl.contract_yaml.contract_yaml_source.resolve_on_read_soda_variable_values
        )

    def _calculate_freshness(self, max_timestamp: datetime, data_timestamp: datetime) -> timedelta:
        return data_timestamp - max_timestamp

    def _freshness_to_seconds(self, freshness: Optional[timedelta]) -> int:
        return floor(freshness.total_seconds())

    def _convert_freshness_seconds_to_check_unit(self, freshness_in_seconds: int, unit: str) -> float:
        if unit == "second":
            threshold_value = freshness_in_seconds
        elif unit == "minute":
            threshold_value = freshness_in_seconds / 60
        elif unit == "hour":
            threshold_value = freshness_in_seconds / (60 * 60)
        elif unit == "day":
            threshold_value = freshness_in_seconds / (60 * 60 * 24)
        else:
            raise ValueError(f"Unknown time unit: {unit}")

        return threshold_value

    def _get_max_timestamp(
        self, measurement_values: MeasurementValues, metric: MetricImpl, column_name: str
    ) -> Optional[datetime]:
        max_timestamp: Optional[datetime] = measurement_values.get_value(metric)
        if max_timestamp is None:
            logger.warning(
                f"Freshness metric '{metric.type}' for column '{column_name}' "
                f"returned no value. Does the table or partition have rows?"
            )
            return None
        elif not isinstance(max_timestamp, datetime):
            logger.debug(
                f"Attempting to convert freshness value '{max_timestamp}' of data type '{type(max_timestamp).__name__}' to datetime"
            )
            if isinstance(max_timestamp, date):
                max_timestamp = datetime.combine(max_timestamp, datetime.min.time())
            elif isinstance(max_timestamp, str):
                max_timestamp = convert_str_to_datetime(max_timestamp)

        if not isinstance(max_timestamp, datetime):
            logger.error(
                f"Freshness column '{column_name}' returned value '{max_timestamp}' of data type '{type(max_timestamp).__name__}' which is not a datetime or datetime-compatible type."
            )
            max_timestamp = None

        return max_timestamp

    def _get_max_timestamp_utc(self, max_timestamp: Optional[datetime]) -> Optional[datetime]:
        return self._datetime_to_utc(max_timestamp) if isinstance(max_timestamp, datetime) else None

    def _get_now_timestamp(self) -> Optional[datetime]:
        if self.now_variable is None:
            return self.contract_impl.contract_yaml.data_timestamp
        else:
            now_timestamp_str: str = self.resolved_variable_values.get(self.now_variable)
            if now_timestamp_str is None:
                logger.error(f"Freshness variable '{self.now_variable}' not available")
                return None
            if not isinstance(now_timestamp_str, str):
                logger.error(
                    f"Freshness variable '{self.now_variable}' has wrong " f"type: {type(now_timestamp_str).__name__}"
                )
                return None
            else:
                now_timestamp: Optional[datetime] = convert_str_to_datetime(now_timestamp_str)
                if not isinstance(now_timestamp, datetime):
                    logger.error(f"Freshness variable '{self.now_variable}' is not a timestamp: {now_timestamp_str}")
                return now_timestamp

    def _get_now_timestamp_utc(self, now_timestamp: Optional[datetime]) -> Optional[datetime]:
        return self._datetime_to_utc(now_timestamp) if now_timestamp else None

    @staticmethod
    def _datetime_to_utc(input: datetime) -> datetime:
        if input.tzinfo is None:
            return input.replace(tzinfo=timezone.utc)
        return input.astimezone(timezone.utc)


class FreshnessCheckImpl(FreshnessCheckImplBase):
    def __init__(
        self,
        contract_impl: ContractImpl,
        column_impl: ColumnImpl,
        check_yaml: FreshnessCheckYaml,
    ):
        super().__init__(
            contract_impl=contract_impl,
            column_impl=column_impl,
            check_yaml=check_yaml,
        )
        self.column = check_yaml.column
        self.threshold = ThresholdImpl.create(
            threshold_yaml=check_yaml.threshold,
        )

    def setup_metrics(
        self,
        contract_impl: ContractImpl,
        column_impl: ColumnImpl,
        check_yaml: FreshnessCheckYaml,
    ):
        self.max_timestamp_metric = self._resolve_metric(
            MaxTimestampMetricImpl(
                contract_impl=contract_impl,
                check_impl=self,
                column=self.column,
                now_variable=self.now_variable,
                unit=self.unit,
            )
        )
        self.check_rows_tested_metric_impl: MetricImpl = self._resolve_metric(
            RowCountMetricImpl(contract_impl=contract_impl, check_impl=self)
        )

    def evaluate(self, measurement_values: MeasurementValues) -> CheckResult:
        outcome: CheckOutcome = CheckOutcome.NOT_EVALUATED

        max_timestamp: Optional[datetime] = self._get_max_timestamp(
            measurement_values, self.max_timestamp_metric, self.column
        )
        max_timestamp_utc: Optional[datetime] = self._get_max_timestamp_utc(max_timestamp)
        data_timestamp: Optional[datetime] = self._get_now_timestamp()
        data_timestamp_utc: Optional[datetime] = self._get_now_timestamp_utc(data_timestamp)

        check_rows_tested: int = measurement_values.get_value(self.check_rows_tested_metric_impl)
        diagnostic_metric_values: dict[str, float] = {
            "dataset_rows_tested": self.contract_impl.dataset_rows_tested,
            "check_rows_tested": check_rows_tested,
        }
        freshness: Optional[timedelta] = None
        freshness_in_seconds: Optional[int] = None
        threshold_metric_name: str = f"freshness_in_{self.unit}s"

        threshold_value: Optional[float] = None
        if max_timestamp_utc is None or data_timestamp_utc is None:
            outcome = CheckOutcome.FAILED
        else:
            logger.debug(
                f"Calculating freshness using '{max_timestamp}' as 'max' and '{data_timestamp}' as 'now' values"
            )
            freshness = self._calculate_freshness(max_timestamp_utc, data_timestamp_utc)
            freshness_in_seconds = self._freshness_to_seconds(freshness)

            threshold_value = self._convert_freshness_seconds_to_check_unit(freshness_in_seconds, self.unit)
            diagnostic_metric_values[threshold_metric_name] = threshold_value

            outcome = self.evaluate_threshold(threshold_value)

        freshness_str: Optional[str] = str(freshness) if freshness is not None else None

        return FreshnessCheckResult(
            check=self._build_check_info(),
            outcome=outcome,
            threshold_value=threshold_value,
            diagnostic_metric_values=diagnostic_metric_values,
            max_timestamp=max_timestamp,
            max_timestamp_utc=max_timestamp_utc,
            data_timestamp=data_timestamp,
            data_timestamp_utc=data_timestamp_utc,
            freshness=freshness_str,
            freshness_in_seconds=freshness_in_seconds,
            unit=self.unit,
        )


class MaxTimestampMetricImpl(AggregationMetricImpl):
    def __init__(
        self,
        contract_impl: ContractImpl,
        check_impl: FreshnessCheckImpl,
        column: str,
        now_variable: Optional[str],
        unit: Optional[str],
        data_source_impl: Optional[DataSourceImpl] = None,
        dataset_identifier: Optional[DatasetIdentifier] = None,
    ):
        self.column: str = column
        self.now_variable: Optional[str] = now_variable
        self.unit: Optional[str] = unit
        super().__init__(
            contract_impl=contract_impl,
            metric_type=check_impl.type,
            check_filter=check_impl.check_yaml.filter,
            data_source_impl=data_source_impl,
            dataset_identifier=dataset_identifier,
        )

    def _get_id_properties(self) -> dict[str, any]:
        id_properties: dict[str, str] = super()._get_id_properties()
        id_properties["column"] = self.column
        id_properties["now_variable"] = self.now_variable
        id_properties["unit"] = self.unit
        return id_properties

    def sql_expression(self) -> SqlExpression:
        max_expression = self.column

        if self.check_filter:
            max_expression = CASE_WHEN(SqlExpressionStr(self.check_filter), self.column)

        return MAX(max_expression)

    def convert_db_value(self, value) -> any:
        return value


class FreshnessCheckResult(CheckResult):
    def __init__(
        self,
        check: Check,
        outcome: CheckOutcome,
        threshold_value: Optional[float | int],
        diagnostic_metric_values: Optional[dict[str, float]],
        max_timestamp: Optional[datetime],
        max_timestamp_utc: Optional[datetime],
        data_timestamp: datetime,
        data_timestamp_utc: datetime,
        freshness: Optional[str],
        freshness_in_seconds: Optional[int],
        unit: Optional[str],
    ):
        super().__init__(
            check=check,
            outcome=outcome,
            threshold_value=threshold_value,
            diagnostic_metric_values=diagnostic_metric_values,
        )
        self.max_timestamp: Optional[datetime] = max_timestamp
        self.max_timestamp_utc: Optional[datetime] = max_timestamp_utc
        self.data_timestamp: datetime = data_timestamp
        self.data_timestamp_utc: datetime = data_timestamp_utc
        self.freshness: Optional[str] = freshness
        self.freshness_in_seconds: Optional[int] = freshness_in_seconds
        self.unit: Optional[str] = unit

    # def log_summary(self, logs: Logs) -> None:
    #     super().log_summary(logs)
    #
    #     if self.max_timestamp is None:
    #         logger.info("  Max timestamp has no value. Did the table or partition have rows?")
    #     elif isinstance(self.max_timestamp, datetime):
    #         logger.debug(f"  Max timestamp was {self.max_timestamp}")
    #     else:
    #         logger.error(
    #             f"  Invalid data type for max timestamp ({type(self.max_timestamp).__name__}). "
    #             f"Is the column a timestamp?"
    #         )
    #     if isinstance(self.max_timestamp_utc, datetime):
    #         logger.info(f"  Max timestamp in UTC was {self.max_timestamp_utc}")
    #     if isinstance(self.data_timestamp, datetime):
    #         logger.debug(f"  Data timestamp was {self.data_timestamp}")
    #     if isinstance(self.data_timestamp_utc, datetime):
    #         logger.info(f"  Data timestamp in UTC was {self.data_timestamp_utc}")
    #
    #     if isinstance(self.freshness, str):
    #         logger.debug(f"  Freshness was {self.freshness}")
    #     if isinstance(self.freshness, str):
    #         logger.debug(f"  Freshness in seconds was {self.freshness_in_seconds}")
    #     if isinstance(self.unit, str):
    #         logger.debug(f"  Unit was {self.unit}")



import abc
from typing import Literal

from pydantic import BaseModel, ConfigDict, Field
from soda_core.model.data_source.data_source_connection_properties import (
    DataSourceConnectionProperties,
)


class DataSourceBase(BaseModel, abc.ABC):
    model_config = ConfigDict(
        frozen=True,
        extra="forbid",
        validate_by_name=True,  # Allow to use both field names and aliases when populating from dict
    )
    name: str = Field(..., description="Data source name")
    type: Literal["data_source"]
    # Alias connection -> connection_properties to read that from yaml config
    connection_properties: DataSourceConnectionProperties = Field(
        ..., alias="connection", description="Data source connection details"
    )

    @classmethod
    def get_class_type(cls) -> str:
        """The only way to get the simple type like 'postgres' for plugin registry before instantiating the class"""
        return cls.model_fields["type"].default



import logging
from typing import Dict, Optional, Union

from soda_core.common.data_source_impl import DataSourceImpl
from soda_core.common.exceptions import (
    InvalidArgumentException,
    InvalidDataSourceConfigurationException,
    SodaCloudException,
)
from soda_core.common.logging_constants import soda_logger
from soda_core.common.soda_cloud import SodaCloud
from soda_core.common.yaml import ContractYamlSource, DataSourceYamlSource
from soda_core.contracts.contract_verification import (
    ContractVerificationSession,
    ContractVerificationSessionResult,
)
from soda_core.telemetry.soda_telemetry import SodaTelemetry
from typing_extensions import deprecated

logger: logging.Logger = soda_logger

soda_telemetry = SodaTelemetry()

__AT_LEAST_ONE_CONTRACT_OR_DATASET_REQUIRED = "At least one of -c/--contract or -d/--dataset arguments is required."


@deprecated("Use verify_contract_locally instead")
def verify_contracts_locally(
    data_source_file_path: Optional[str] = None,
    data_source_file_paths: list[str] = [],
    data_sources: Optional[Union[list[DataSourceImpl], DataSourceImpl]] = [],
    contract_file_paths: Optional[Union[str, list[str]]] = None,
    dataset_identifiers: Optional[list[str]] = None,
    soda_cloud_file_path: Optional[str] = None,
    variables: Optional[Dict[str, str]] = None,
    data_timestamp: Optional[str] = None,
    publish: bool = False,
    check_paths: Optional[list[str]] = None,
    dwh_data_source_file_path: Optional[str] = None,
) -> ContractVerificationSessionResult:
    if not contract_file_paths and not dataset_identifiers:
        raise InvalidArgumentException(__AT_LEAST_ONE_CONTRACT_OR_DATASET_REQUIRED)

    if isinstance(contract_file_paths, str):
        contract_file_paths = [contract_file_paths]

    if (
        contract_file_paths and len(contract_file_paths) > 1
    ):  # We still need the "none" check, if there is no contract file path, we don't want to raise an error for the len()
        raise InvalidArgumentException("Only one contract is allowed at a time")

    if dataset_identifiers and len(dataset_identifiers) > 1:
        raise InvalidArgumentException("Only one dataset identifier is allowed at a time")

    if contract_file_paths and dataset_identifiers:
        logger.info(
            "Both contract file paths and dataset identifiers are provided. Only evaluating the contract file paths."
        )

    assert isinstance(dataset_identifiers, list), f"Expected a list, got {type(dataset_identifiers)}"

    contract_file_path = __attempt_pick_first_element(contract_file_paths)
    dataset_identifier = __attempt_pick_first_element(dataset_identifiers)

    return verify_contract_locally(
        data_source_file_path=data_source_file_path,
        data_source_file_paths=data_source_file_paths,
        data_sources=data_sources,
        contract_file_path=contract_file_path,
        dataset_identifier=dataset_identifier,
        soda_cloud_file_path=soda_cloud_file_path,
        variables=variables,
        data_timestamp=data_timestamp,
        publish=publish,
        check_paths=check_paths,
        dwh_data_source_file_path=dwh_data_source_file_path,
    )


def verify_contract_locally(
    data_source_file_path: Optional[str] = None,
    data_source_file_paths: list[str] = [],
    data_sources: Optional[Union[list[DataSourceImpl], DataSourceImpl]] = [],
    contract_file_path: Optional[str] = None,
    dataset_identifier: Optional[str] = None,
    soda_cloud_file_path: Optional[str] = None,
    variables: Optional[Dict[str, str]] = None,
    data_timestamp: Optional[str] = None,
    publish: bool = False,
    check_paths: Optional[list[str]] = None,
    dwh_data_source_file_path: Optional[str] = None,
) -> ContractVerificationSessionResult:
    """
    Verifies the contract locally.
    """
    if not isinstance(data_sources, list):
        data_sources = [data_sources]

    return verify_contract(
        contract_file_path=contract_file_path,
        dataset_identifier=dataset_identifier,
        data_source_file_path=data_source_file_path,
        data_source_file_paths=data_source_file_paths,
        data_sources=data_sources,
        soda_cloud_file_path=soda_cloud_file_path,
        variables=variables,
        data_timestamp=data_timestamp,
        publish=publish,
        use_agent=False,
        check_paths=check_paths,
        dwh_data_source_file_path=dwh_data_source_file_path,
    )


@deprecated("Use verify_contract_on_agent instead")
def verify_contracts_on_agent(
    soda_cloud_file_path: str,
    contract_file_paths: Optional[Union[str, list[str]]] = None,
    dataset_identifiers: Optional[list[str]] = None,
    data_source_file_path: Optional[str] = None,
    data_source_file_paths: list[str] = [],
    variables: Optional[Dict[str, str]] = None,
    publish: bool = False,
    verbose: bool = False,
    blocking_timeout_in_minutes: int = 60,
) -> ContractVerificationSessionResult:
    if not contract_file_paths and not dataset_identifiers:
        raise InvalidArgumentException(__AT_LEAST_ONE_CONTRACT_OR_DATASET_REQUIRED)

    if isinstance(contract_file_paths, str):
        contract_file_paths = [contract_file_paths]

    if (
        contract_file_paths and len(contract_file_paths) > 1
    ):  # We still need the "none" check, if there is no contract file path, we don't want to raise an error for the len()
        raise InvalidArgumentException("Only one contract is allowed at a time")

    if dataset_identifiers and len(dataset_identifiers) > 1:
        raise InvalidArgumentException("Only one dataset identifier is allowed at a time")

    if contract_file_paths and dataset_identifiers:
        logger.info(
            "Both contract file paths and dataset identifiers are provided. Only evaluating the contract file paths."
        )

    assert isinstance(dataset_identifiers, list), f"Expected a list, got {type(dataset_identifiers)}"

    contract_file_path = __attempt_pick_first_element(contract_file_paths)
    dataset_identifier = __attempt_pick_first_element(dataset_identifiers)

    return verify_contract_on_agent(
        soda_cloud_file_path=soda_cloud_file_path,
        contract_file_path=contract_file_path,
        dataset_identifier=dataset_identifier,
        data_source_file_path=data_source_file_path,
        data_source_file_paths=data_source_file_paths,
        variables=variables,
        publish=publish,
        verbose=verbose,
        blocking_timeout_in_minutes=blocking_timeout_in_minutes,
    )


def verify_contract_on_agent(
    soda_cloud_file_path: str,
    contract_file_path: Optional[str] = None,
    dataset_identifier: Optional[str] = None,
    data_source_file_path: Optional[str] = None,
    data_source_file_paths: list[str] = [],
    variables: Optional[Dict[str, str]] = None,
    publish: bool = False,
    verbose: bool = False,
    blocking_timeout_in_minutes: int = 60,
) -> ContractVerificationSessionResult:
    """
    Verifies the contract on an agent.
    """
    return verify_contract(
        contract_file_path=contract_file_path,
        dataset_identifier=dataset_identifier,
        data_source_file_path=data_source_file_path,
        data_source_file_paths=data_source_file_paths,
        soda_cloud_file_path=soda_cloud_file_path,
        variables=variables,
        publish=publish,
        verbose=verbose,
        use_agent=True,
        blocking_timeout_in_minutes=blocking_timeout_in_minutes,
    )


def verify_contract(
    contract_file_path: Optional[str],
    dataset_identifier: Optional[str],
    data_source_file_path: Optional[str],
    soda_cloud_file_path: Optional[str],
    publish: bool,
    use_agent: bool,
    variables: Optional[Dict[str, str]] = None,
    data_timestamp: Optional[str] = None,
    verbose: bool = False,
    blocking_timeout_in_minutes: int = 60,
    data_sources: Optional[list[DataSourceImpl]] = None,
    data_source_file_paths: Optional[list[str]] = None,
    check_paths: Optional[list[str]] = None,
    dwh_data_source_file_path: Optional[str] = None,
) -> ContractVerificationSessionResult:
    if not data_source_file_paths:
        data_source_file_paths = []

    # Backward compatibility for single data source file path - append to the list of data source file paths
    if data_source_file_path and data_source_file_path not in data_source_file_paths:
        data_source_file_paths.append(data_source_file_path)

    # Backward compatibility for single dataset identifier - convert to list
    if dataset_identifier is not None:
        dataset_identifiers = [dataset_identifier]
    else:
        dataset_identifiers = None

    # TODO: change this when we fully deprecate a list of contract file paths
    # This is only here to make sure the rest of the code still works.
    if isinstance(contract_file_path, str):
        contract_file_paths = [contract_file_path]
    else:
        contract_file_paths = None

    soda_cloud_client: Optional[SodaCloud] = None
    try:
        if soda_cloud_file_path:
            soda_cloud_client = SodaCloud.from_config(soda_cloud_file_path, variables)

        # TODO: verify the path where connection is provided
        validate_verify_arguments(
            contract_file_paths,
            dataset_identifiers,
            data_source_file_paths,
            data_sources,
            publish,
            use_agent,
            soda_cloud_client,
        )

        contract_yaml_sources = _create_contract_yamls(contract_file_paths, dataset_identifiers, soda_cloud_client)

        if len(contract_yaml_sources) == 0:
            soda_logger.debug("No contracts given. Exiting.")
            return ContractVerificationSessionResult(contract_verification_results=[])

        data_source_yaml_sources: list[DataSourceYamlSource] = []
        if data_source_file_paths or dataset_identifiers:
            data_source_yaml_sources.extend(
                _create_datasource_yamls(
                    data_source_file_paths,
                    dataset_identifiers,
                    soda_cloud_client,
                    use_agent,
                )
            )

        contract_verification_result = ContractVerificationSession.execute(
            contract_yaml_sources=contract_yaml_sources,
            data_source_yaml_sources=data_source_yaml_sources,
            data_source_impls=data_sources,
            soda_cloud_impl=soda_cloud_client,
            variables=variables,
            data_timestamp=data_timestamp,
            only_validate_without_execute=False,
            soda_cloud_publish_results=publish,
            soda_cloud_use_agent=use_agent,
            soda_cloud_verbose=verbose,
            soda_cloud_use_agent_blocking_timeout_in_minutes=blocking_timeout_in_minutes,
            check_paths=check_paths,
            dwh_data_source_file_path=dwh_data_source_file_path,
        )

        soda_telemetry.ingest_contract_verification_session_result(
            contract_verification_session_result=contract_verification_result
        )

        return contract_verification_result
    except Exception as exc:
        soda_logger.error(exc)
        if soda_cloud_client:
            soda_cloud_client.mark_scan_as_failed(exc=exc)
        raise exc


def validate_verify_arguments(
    contract_file_paths: Optional[list[str]],
    dataset_identifiers: Optional[list[str]],
    data_source_file_paths: Optional[list[str]],
    data_sources: Optional[list[DataSourceImpl]],
    publish: bool,
    use_agent: bool,
    soda_cloud_client: Optional[SodaCloud],
) -> None:
    if publish and not soda_cloud_client:
        raise InvalidArgumentException(
            "A Soda Cloud configuration file is required to use the -p/--publish argument. "
            "Please provide the '--soda-cloud' argument with a valid configuration file path."
        )

    if use_agent and not soda_cloud_client:
        raise InvalidArgumentException(
            "A Soda Cloud configuration file is required to use the -a/--agent argument. "
            "Please provide the '--soda-cloud' argument with a valid configuration file path."
        )

    if all_none_or_empty(contract_file_paths, dataset_identifiers):
        raise InvalidArgumentException(__AT_LEAST_ONE_CONTRACT_OR_DATASET_REQUIRED)

    if dataset_identifiers and not soda_cloud_client:
        raise InvalidArgumentException(
            "A Soda Cloud configuration file is required to use the -d/--dataset argument."
            "Please provide the '--soda-cloud' argument with a valid configuration file path."
        )

    if all_none_or_empty(dataset_identifiers) and not data_source_file_paths and not use_agent and not data_sources:
        raise InvalidArgumentException("At least one of -ds/--data-source or -d/--dataset value is required.")


def all_none_or_empty(*args: list | None) -> bool:
    return all(x is None or len(x) == 0 for x in args)


def is_using_remote_contract(
    contract_file_paths: Optional[list[str]], dataset_identifiers: Optional[list[str]]
) -> bool:
    return (contract_file_paths is None or len(contract_file_paths) == 0) and dataset_identifiers is not None


def is_using_remote_datasource(dataset_identifiers: Optional[list[str]], data_source_file_path: Optional[str]) -> bool:
    return not data_source_file_path and not all_none_or_empty(dataset_identifiers)


def contract_verification_is_not_sent_to_cloud(
    contract_verification_session_result: ContractVerificationSessionResult,
) -> bool:
    return any(
        cr.sending_results_to_soda_cloud_failed
        for cr in contract_verification_session_result.contract_verification_results
    )


def _create_contract_yamls(
    contract_file_paths: Optional[list[str]],
    dataset_identifiers: Optional[list[str]],
    soda_cloud_client: SodaCloud,
) -> list[ContractYamlSource]:
    contract_yaml_sources = []

    if contract_file_paths:
        contract_yaml_sources += [ContractYamlSource.from_file_path(p) for p in contract_file_paths]

    if is_using_remote_contract(contract_file_paths, dataset_identifiers) and soda_cloud_client:
        for dataset_identifier in dataset_identifiers:
            try:
                contract = soda_cloud_client.fetch_contract_for_dataset(dataset_identifier)
                contract_yaml_sources.append(ContractYamlSource.from_str(contract))
            except SodaCloudException as exc:
                soda_logger.error(f"Could not fetch contract for dataset '{dataset_identifier}': skipping verification")

    if not contract_yaml_sources:
        return []

    return contract_yaml_sources


def _create_datasource_yamls(
    data_source_file_paths: Optional[list[str]],
    dataset_identifiers: Optional[list[str]],
    soda_cloud_client: SodaCloud,
    use_agent: bool,
) -> Optional[DataSourceYamlSource]:
    data_source_yamls: list[DataSourceYamlSource] = []

    if data_source_file_paths:
        for data_source_file_path in data_source_file_paths:
            soda_logger.debug(f"Using local data source config: {data_source_file_path}")
            data_source_yamls.append(DataSourceYamlSource.from_file_path(data_source_file_path))

    if dataset_identifiers:
        for dataset_identifier in dataset_identifiers:
            soda_logger.debug(
                f"No local data source config for '{dataset_identifier}', trying to fetch data source config from cloud"
            )
            data_source_config = soda_cloud_client.fetch_data_source_configuration_for_dataset(dataset_identifier)

            if not data_source_config:
                soda_logger.error(
                    f"Could not fetch data source configuration for dataset '{dataset_identifier}': skipping fetch."
                )
            else:
                data_source_yamls.append(DataSourceYamlSource.from_str(data_source_config))

    if data_source_yamls:
        return data_source_yamls

    # By this point, we can only progress if we are using an agent.
    # Then the agent will provide the data source config.
    if use_agent:
        return None

    raise InvalidDataSourceConfigurationException(
        "No data source configuration provided and no dataset identifiers given. "
        "Please provide a data source configuration file or a dataset identifier."
    )


def __attempt_pick_first_element(my_list: Optional[list[str]]) -> Optional[str]:
    if my_list is None:
        return None
    if isinstance(my_list, list) and len(my_list) >= 1:
        return my_list[0]
    raise InvalidArgumentException(
        "Expected a list with at least one element, got an empty list or another type of object."
    )



# from __future__ import annotations
#
# import json
#
# from jsonschema.validators import Draft7Validator
# from soda_core.data_source.logs import Logs
#
#
# class ValidatorLoader:
#
#     @classmethod
#     def load_json_schema_validator(cls) -> Draft7Validator:
#         suffix = "/impl/json_schema_verifier.py"
#         contracts_dir = __file__[: -len(suffix)]
#         contract_schema_json_file_path = f"{contracts_dir}/soda_data_contract_json_schema_1_0_0.json"
#         with open(contract_schema_json_file_path) as f:
#             contract_schema_json_str = f.read()
#             schema_dict = json.loads(contract_schema_json_str)
#             return Draft7Validator(schema_dict)
#
#
# class JsonSchemaVerifier:
#
#     __validator = ValidatorLoader.load_json_schema_validator()
#
#     def __init__(self, logs: Optional[Logs] = None):
#         # See also adr/03_exceptions_vs_error_logs.md
#         self.logs: Logs = logs if logs else Logs()
#
#     def verify(self, yaml_object: object) -> None:
#         """
#         Verifies that the YAML data structure matches the data contract schema.
#         Swallows all errors and exceptions and appends them to self.logs.
#         """
#         for error in self.__validator.iter_errors(instance=yaml_object):
#             error_path_text = "contract document level" if len(error.path) == 0 else error.json_path
#             self.logs.error(f"JSON schema error: {error.message} ({error_path_text})")



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Callable, Dict, Optional, Type

from soda_core.common.data_source_connection import DataSourceConnection
from soda_core.common.data_source_results import (
    QueryResult,
    QueryResultIterator,
    UpdateResult,
)
from soda_core.common.exceptions import DataSourceConnectionException
from soda_core.common.logging_constants import soda_logger
from soda_core.common.metadata_types import (
    ColumnMetadata,
    DataSourceNamespace,
    DbSchemaDataSourceNamespace,
    SchemaDataSourceNamespace,
)
from soda_core.common.sql_dialect import SqlDialect
from soda_core.common.statements.metadata_tables_query import (
    FullyQualifiedTableName,
    MetadataTablesQuery,
)
from soda_core.common.statements.table_types import FullyQualifiedObjectName, TableType
from soda_core.common.yaml import DataSourceYamlSource, YamlObject
from soda_core.contracts.contract_verification import DataSource
from soda_core.model.data_source.data_source import DataSourceBase

logger: logging.Logger = soda_logger

if TYPE_CHECKING:
    from soda_core.contracts.impl.contract_verification_impl import ContractImpl


class DataSourceImpl(ABC):
    __implementation_classes: Dict[str, Callable[[], Type["DataSourceImpl"]]] = {}
    __model_classes: Dict[str, Type[DataSourceBase]] = {}

    @classmethod
    def from_yaml_source(
        cls, data_source_yaml_source: DataSourceYamlSource, provided_variable_values: Optional[dict] = None
    ) -> Optional[DataSourceImpl]:
        data_source_yaml_source.resolve(variables=provided_variable_values)
        data_source_yaml: YamlObject = data_source_yaml_source.parse()
        if not data_source_yaml:
            return None

        type_name = data_source_yaml.yaml_dict.get("type")
        if not type_name:
            raise ValueError("Missing required 'type' in data source YAML")

        impl_class = cls.__implementation_classes.get(type_name)
        if not impl_class:
            raise ImportError(
                f"Data source type '{type_name}' not available. "
                f"Make sure to install the required plugin, e.g. `pip install soda-{type_name}`"
            )

        model_class = cls.__model_classes.get(type_name)
        if not model_class:
            raise ImportError(
                f"Model class for data source type '{type_name}' not found. "
                f"This is likely a bug in the plugin implementation."
            )

        validated_model = model_class.model_validate(data_source_yaml.yaml_dict)
        return impl_class(data_source_model=validated_model)

    def __init__(
        self,
        data_source_model: DataSourceBase,
        connection: Optional[DataSourceConnection] = None,
    ):
        self.data_source_model: DataSourceBase = data_source_model
        self.name: str = data_source_model.name
        self.type_name: str = data_source_model.get_class_type()
        self.sql_dialect: SqlDialect = self._create_sql_dialect()
        self.data_source_connection: Optional[DataSourceConnection] = connection

    def __init_subclass__(cls, model_class: Type[DataSourceBase], **kwargs):
        super().__init_subclass__(**kwargs)
        type_name = model_class.get_class_type()
        cls.__model_classes[type_name] = model_class
        cls.__implementation_classes[type_name] = cls

    def create_copy_with_different_connection(self, connection: DataSourceConnection) -> DataSourceImpl:
        "Use this method when you want to use the DataSourceImpl while having multiple connections open. Can be used for bulk inserts for example."
        return self.__class__(data_source_model=self.data_source_model, connection=connection)

    def __str__(self) -> str:
        return self.name

    @abstractmethod
    def _create_data_source_connection(self) -> DataSourceConnection:
        pass

    @abstractmethod
    def _create_sql_dialect(self) -> SqlDialect:
        pass

    def __enter__(self) -> None:
        return self.connection

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_connection()

    @property
    def connection(self) -> DataSourceConnection:
        if not self.has_open_connection():
            self.open_connection()

        return self.data_source_connection

    def open_connection(self) -> None:
        try:
            self.data_source_connection = self._create_data_source_connection()
        except Exception as e:
            raise DataSourceConnectionException(e) from e

    def has_open_connection(self) -> bool:
        return (
            isinstance(self.data_source_connection, DataSourceConnection)
            and self.data_source_connection.connection is not None
        )

    def close_connection(self) -> None:
        if self.has_open_connection():
            self.data_source_connection.close_connection()

    def execute_query(self, sql: str, log_query: bool = True) -> QueryResult:
        return self.connection.execute_query(sql=sql, log_query=log_query)

    def execute_query_one_by_one(
        self,
        sql: str,
        row_callback: Callable[[tuple, tuple[tuple]], None],
        log_query: bool = True,
        row_limit: Optional[int] = None,
    ) -> tuple[tuple]:
        # Returns the description of the query.
        return self.data_source_connection.execute_query_one_by_one(
            sql=sql, row_callback=row_callback, log_query=log_query, row_limit=row_limit
        )

    def execute_query_iterate(self, sql: str, log_query: bool = True) -> QueryResultIterator:
        return self.data_source_connection.execute_query_iterate(sql=sql, log_query=log_query)

    def execute_update(self, sql: str, log_query: bool = True) -> UpdateResult:
        return self.connection.execute_update(sql=sql, log_query=log_query)

    def test_connection_error_message(self) -> Optional[str]:
        try:
            with self:
                self.data_source_connection.execute_query(f"SELECT 1")
                return None
        except Exception as e:
            return str(e)

    def build_data_source(self) -> DataSource:
        return DataSource(name=self.name, type=self.type_name)

    # TODO refactor to method here and delegate query building and result extraction to SqlDialect similar to get_columns_metadata
    def create_metadata_tables_query(self) -> MetadataTablesQuery:
        return MetadataTablesQuery(sql_dialect=self.sql_dialect, data_source_connection=self.data_source_connection)

    def parse_column_names_from_query_result(self, query_result: QueryResult) -> list[str]:
        return [self.connection._execute_query_get_result_row_column_name(column) for column in query_result.columns]

    def get_columns_metadata(self, dataset_prefixes: list[str], dataset_name: str) -> list[ColumnMetadata]:
        sql: str = self.build_columns_metadata_query_str(dataset_prefixes=dataset_prefixes, dataset_name=dataset_name)
        query_result: QueryResult = self.execute_query(sql)
        return self.sql_dialect.build_column_metadatas_from_query_result(query_result)

    def build_columns_metadata_query_str(self, dataset_prefixes: list[str], dataset_name: str) -> str:
        schema_name: Optional[str] = self.extract_schema_from_prefix(dataset_prefixes)
        database_name: Optional[str] = self.extract_database_from_prefix(dataset_prefixes)

        table_namespace: DataSourceNamespace = (
            SchemaDataSourceNamespace(schema=schema_name)
            if database_name is None
            else DbSchemaDataSourceNamespace(database=database_name, schema=schema_name)
        )

        # BigQuery must be able to override to get the location
        return self.sql_dialect.build_columns_metadata_query_str(
            table_namespace=table_namespace, table_name=dataset_name
        )

    def extract_schema_from_prefix(self, prefixes: list[str]) -> Optional[str]:
        schema_index: int | None = self.sql_dialect.get_schema_prefix_index()
        if schema_index is None:
            return None
        schema_name: str = prefixes[schema_index] if schema_index < len(prefixes) else None
        return schema_name

    def extract_database_from_prefix(self, prefixes: list[str]) -> Optional[str]:
        database_index: int | None = self.sql_dialect.get_database_prefix_index()
        if database_index is None:
            return None
        database_name: str = prefixes[database_index] if database_index < len(prefixes) else None
        return database_name

    def _build_table_namespace_for_schema_query(self, prefixes: list[str]) -> tuple[DataSourceNamespace, str]:
        """
        Builds the table namespace for the schema query.
        Returns the table namespace and the schema name.
        """
        schema_name: Optional[str] = self.extract_schema_from_prefix(prefixes)
        database_name: str | None = self.extract_database_from_prefix(prefixes)
        if schema_name is None:
            raise ValueError(f"Cannot determine schema name from prefixes: {prefixes}")

        table_namespace: DataSourceNamespace = (
            SchemaDataSourceNamespace(schema=schema_name)
            if database_name is None
            else DbSchemaDataSourceNamespace(
                database=database_name,
                schema=schema_name,
            )
        )
        return table_namespace, schema_name

    def test_schema_exists(self, prefixes: list[str]) -> bool:
        table_namespace, schema_name = self._build_table_namespace_for_schema_query(prefixes=prefixes)

        # Query all schemas to check if the target schema exists
        schemas_query_sql: str = self.sql_dialect.build_schemas_metadata_query_str(
            table_namespace=table_namespace, filter_on_schema_name=schema_name
        )
        query_result: QueryResult = self.execute_query(schemas_query_sql)

        # Check if schema exists in the results
        schema_exists: bool = False
        schema_name_lower: str = schema_name.lower()
        for row in query_result.rows:
            if row[0] and row[0].lower() == schema_name_lower:
                schema_exists = True
                break
        return schema_exists

    def verify_if_table_exists(self, prefixes: list[str], table_name: str) -> bool:
        fully_qualified_object_names: list[FullyQualifiedObjectName] = self._get_fully_qualified_table_names(
            prefixes=prefixes, table_name=table_name
        )
        return any(
            fully_qualified_object_name.get_object_name() == table_name
            for fully_qualified_object_name in fully_qualified_object_names
            if isinstance(fully_qualified_object_name, FullyQualifiedTableName)  # Only check TABLES
        )

    def _get_fully_qualified_table_names(self, prefixes: list[str], table_name: str) -> list[FullyQualifiedObjectName]:
        metadata_tables_query: MetadataTablesQuery = self.create_metadata_tables_query()
        database_name = self.extract_database_from_prefix(prefixes)
        schema_name = self.extract_schema_from_prefix(prefixes)

        fully_qualified_object_names: list[FullyQualifiedObjectName] = metadata_tables_query.execute(
            database_name=database_name,
            schema_name=schema_name,
            include_table_name_like_filters=[table_name],
        )
        return fully_qualified_object_names

    def discover_qualified_objects(
        self,
        prefixes: list[str],
        object_types: Optional[list[TableType]] = None,
    ) -> list[FullyQualifiedObjectName]:
        metadata_tables_query: MetadataTablesQuery = self.create_metadata_tables_query()
        database_name = self.extract_database_from_prefix(prefixes)
        schema_name = self.extract_schema_from_prefix(prefixes)

        fully_qualified_object_names: list[FullyQualifiedObjectName] = metadata_tables_query.execute(
            database_name=database_name,
            schema_name=schema_name,
            types_to_return=object_types,
        )
        return fully_qualified_object_names

    def switch_warehouse(self, warehouse: str, contract_impl: ContractImpl) -> None:
        # Noop by default, only some data sources need to implement this
        pass

    def get_current_warehouse(self) -> Optional[str]:
        # Noop by default, only some data sources need to implement this
        return None



from __future__ import annotations

import logging
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from logging import ERROR, WARNING, LogRecord
from numbers import Number
from typing import Any, Optional

from soda_core import is_verbose
from soda_core.common.logging_constants import Emoticons, soda_logger
from soda_core.common.logs import Location
from soda_core.common.yaml import ContractYamlSource, DataSourceYamlSource
from soda_core.contracts.contract_interfaces import Loggable

logger: logging.Logger = soda_logger


class ContractVerificationSession:
    """Represents the contract verification session.

    Multiple Contracts over multiple Data Sources can be verified in one verification session.

    @param contract_yaml_sources: The list of contract YAML sources to verify.
    @param only_validate_without_execute: If True, only validate the contracts without executing them.
    @param variables: The variables to use in the contract queries.
    @param data_timestamp: The timestamp of the data to use for the verification.
    @param data_source_impls: The data source implementations to use for the verification.
    @param data_source_yaml_sources: The data source YAML sources to use for the verification.
    @param soda_cloud_impl: The Soda Cloud implementation to use for the verification.
    @param soda_cloud_publish_results: If True, publish the results to Soda Cloud.
    @param soda_cloud_use_agent: If True, use the Soda Cloud agent for the verification.
    @param soda_cloud_verbose: If True, enable verbose logging for the Soda Cloud agent.
    """

    @classmethod
    def execute(
        cls,
        contract_yaml_sources: list[ContractYamlSource],
        only_validate_without_execute: bool = False,
        variables: Optional[dict[str, str]] = None,
        data_timestamp: Optional[str] = None,
        data_source_impls: Optional[list["DataSourceImpl"]] = None,
        data_source_yaml_sources: Optional[list[DataSourceYamlSource]] = None,
        soda_cloud_impl: Optional["SodaCloud"] = None,
        soda_cloud_publish_results: bool = False,
        soda_cloud_use_agent: bool = False,
        soda_cloud_verbose: bool = False,
        soda_cloud_use_agent_blocking_timeout_in_minutes: int = 60,
        check_paths: Optional[list[str]] = None,
        dwh_data_source_file_path: Optional[str] = None,
    ) -> ContractVerificationSessionResult:
        from soda_core.contracts.impl.contract_verification_impl import (
            ContractVerificationSessionImpl,
        )

        return ContractVerificationSessionImpl.execute(
            contract_yaml_sources=contract_yaml_sources,
            only_validate_without_execute=only_validate_without_execute,
            variables=variables,
            data_timestamp=data_timestamp,
            data_source_impls=data_source_impls,
            data_source_yaml_sources=data_source_yaml_sources,
            soda_cloud_impl=soda_cloud_impl,
            soda_cloud_publish_results=soda_cloud_publish_results,
            soda_cloud_use_agent=soda_cloud_use_agent,
            soda_cloud_verbose=soda_cloud_verbose,
            soda_cloud_use_agent_blocking_timeout_in_minutes=soda_cloud_use_agent_blocking_timeout_in_minutes,
            check_paths=check_paths,
            dwh_data_source_file_path=dwh_data_source_file_path,
        )


class ContractVerificationSessionResult:
    """Represents the result of a contract verification session.

    Provides overview of logs, errors, and the status of the verification process over all of the verified Contracts.

    @param contract_verification_results: The list of contract verification results.
    """

    def __init__(self, contract_verification_results: list[ContractVerificationResult]):
        self.contract_verification_results: list[ContractVerificationResult] = contract_verification_results

    def get_logs(self) -> list[str]:
        logs: list[str] = []
        for contract_verification_result in self.contract_verification_results:
            logs.extend(contract_verification_result.get_logs())
        return logs

    def get_logs_str(self) -> str:
        return "\n".join(self.get_logs())

    def get_errors(self) -> list[str]:
        errors: list[str] = []
        for contract_verification_result in self.contract_verification_results:
            errors.extend(contract_verification_result.get_errors())
        return errors

    @property
    def number_of_checks(self) -> int:
        return sum(
            contract_verification_result.number_of_checks
            for contract_verification_result in self.contract_verification_results
        )

    @property
    def number_of_checks_passed(self) -> int:
        return sum(
            contract_verification_result.number_of_checks_passed
            for contract_verification_result in self.contract_verification_results
        )

    @property
    def number_of_checks_failed(self) -> int:
        return sum(
            contract_verification_result.number_of_checks_failed
            for contract_verification_result in self.contract_verification_results
        )

    @property
    def number_of_checks_excluded(self) -> int:
        return sum(
            contract_verification_result.number_of_checks_excluded
            for contract_verification_result in self.contract_verification_results
        )

    def get_errors_str(self) -> str:
        return "\n".join(self.get_errors())

    @property
    def has_errors(self) -> bool:
        return any(
            contract_verification_result.has_errors
            for contract_verification_result in self.contract_verification_results
        )

    @property
    def is_failed(self) -> bool:
        """
        Returns true if there are checks that have failed.
        False is returned if there are no check results.
        Only looks at check results.
        Ignores execution errors in the logs.
        """
        return any(
            contract_verification_result.is_failed
            for contract_verification_result in self.contract_verification_results
        )

    @property
    def is_warned(self) -> bool:
        """
        Returns true if there are checks that have warnings.
        False is returned if there are no check results.
        Only looks at check results.
        Ignores execution errors in the logs.
        """
        return any(
            contract_verification_result.is_warned
            for contract_verification_result in self.contract_verification_results
        )

    @property
    def is_passed(self) -> bool:
        """
        Returns true if there are no checks that have failed.
        Ignores execution errors in the logs.
        """
        return all(
            contract_verification_result.is_passed
            for contract_verification_result in self.contract_verification_results
        )

    @property
    def has_excluded(self) -> bool:
        """
        Returns true if there are checks that have been excluded.
        Ignores execution errors in the logs.
        """
        return any(
            contract_verification_result.has_excluded
            for contract_verification_result in self.contract_verification_results
        )

    @property
    def is_ok(self) -> bool:
        return all(
            contract_verification_result.is_ok for contract_verification_result in self.contract_verification_results
        )

    def assert_ok(self) -> ContractVerificationSessionResult:
        if not self.is_ok:
            raise SodaException(message=self.get_errors_str())
        return self


class SodaException(Exception):
    """
    See also adr/03_exceptions_vs_error_logs.md
    """

    def __init__(
        self,
        message: Optional[str] = None,
        contract_verification_result: Optional[ContractVerificationSessionResult] = None,
    ):
        self.contract_verification_result: Optional[ContractVerificationSessionResult] = contract_verification_result
        super().__init__(message)


class CheckOutcome(Enum):
    PASSED = "PASSED"
    FAILED = "FAILED"
    WARN = "WARN"
    NOT_EVALUATED = "NOT_EVALUATED"
    EXCLUDED = "EXCLUDED"


@dataclass
class YamlFileContentInfo:
    source_content_str: Optional[str]
    local_file_path: Optional[str]
    git_repo: Optional[str] = None  # Aspirational, not used yet
    soda_cloud_file_id: Optional[str] = None


@dataclass
class Contract:
    data_source_name: Optional[str]
    dataset_prefix: list[str]
    dataset_name: str
    soda_qualified_dataset_name: str
    source: YamlFileContentInfo
    dataset_id: Optional[str] = None  # This one can be filled later when we get the dataset id from Soda Cloud


@dataclass
class DataSource:
    name: str
    type: str


@dataclass
class Threshold:
    level: str
    must_be_greater_than: Optional[Number] = None
    must_be_greater_than_or_equal: Optional[Number] = None
    must_be_less_than: Optional[Number] = None
    must_be_less_than_or_equal: Optional[Number] = None
    must_be: Optional[Number] = None
    must_not_be: Optional[Number] = None

    def __str__(self) -> str:
        parts = []
        if self.must_be is not None:
            parts.append(f"must be: {self.must_be}")
        if self.must_not_be is not None:
            parts.append(f"must not be: {self.must_not_be}")
        if self.must_be_greater_than is not None:
            parts.append(f"must be greater than: {self.must_be_greater_than}")
        if self.must_be_greater_than_or_equal is not None:
            parts.append(f"must be greater than or equal: {self.must_be_greater_than_or_equal}")
        if self.must_be_less_than is not None:
            parts.append(f"must be less than: {self.must_be_less_than}")
        if self.must_be_less_than_or_equal is not None:
            parts.append(f"must be less than or equal: {self.must_be_less_than_or_equal}")
        parts_str = "\n".join(parts)
        return f"level: {self.level}\n{parts_str}" if parts_str else self.level


@dataclass
class Check:
    """Represents the state / essence of a check after it has been executed.

    i.e. "after" CheckImpl has been executed and check result is being created.
    """

    column_name: Optional[str]
    type: str
    qualifier: Optional[str]
    name: Optional[str]
    path: str
    identity: str
    definition: str
    column_name: Optional[str]
    contract_file_line: int
    contract_file_column: int
    threshold: Optional[Threshold]
    attributes: Optional[dict[str, Any]]
    location: Optional[Location]


class CheckResult:
    """
    Represents the result of a check.

    @param check: The check that was performed.
    @param outcome: The outcome of the check.
    @param threshold_value: The threshold value that was applied to the check.
    @param diagnostic_metric_values: The diagnostic metric values collected during the check.
    """

    def __init__(
        self,
        check: Check,
        outcome: CheckOutcome,
        threshold_value: Optional[float | int] = None,
        diagnostic_metric_values: Optional[dict[str, float]] = None,
        autogenerate_diagnostics_payload: bool = False,
    ):
        self.check: Check = check
        self.threshold_value: Optional[float | int] = threshold_value
        self.outcome: CheckOutcome = outcome
        self.diagnostic_metric_values: Optional[dict[str, float | int | str]] = diagnostic_metric_values
        self.autogenerate_diagnostics_payload: bool = autogenerate_diagnostics_payload

    @property
    def outcome_emoticon(self) -> str:
        if self.outcome == CheckOutcome.PASSED:
            return Emoticons.WHITE_CHECK_MARK
        elif self.outcome == CheckOutcome.FAILED:
            return Emoticons.CROSS_MARK
        elif self.outcome == CheckOutcome.WARN:
            return Emoticons.WARNING
        else:
            return Emoticons.QUESTION_MARK

    @property
    def is_passed(self) -> bool:
        return self.outcome == CheckOutcome.PASSED

    @property
    def is_warned(self) -> bool:
        return self.outcome == CheckOutcome.WARN

    @property
    def is_failed(self) -> bool:
        return self.outcome == CheckOutcome.FAILED

    @property
    def is_not_evaluated(self) -> bool:
        return self.outcome == CheckOutcome.NOT_EVALUATED

    def is_excluded(self) -> bool:
        return self.outcome == CheckOutcome.EXCLUDED

    def log_table_row(self) -> dict:
        row = {}
        row["Column"] = self.check.column_name if self.check.column_name else "[dataset-level]"
        row["Check"] = self.check.name
        row["Threshold"] = self.check.threshold
        row["Outcome"] = f"{self.outcome_emoticon} {self.outcome.name}"

        if is_verbose():
            row["Check Type"] = self.check.type
            row["Identity"] = self.check.identity
        row["Diagnostics"] = self.log_table_row_diagnostics(verbose=True if is_verbose() else False)

        return row

    def log_table_row_diagnostics(self, verbose: bool = True) -> str:
        diagnostics = []

        if self.diagnostic_metric_values:
            # If the diagnostic metric values implements the ISodaCloudOutput interface, use it
            if isinstance(self.diagnostic_metric_values, Loggable):
                diagnostics.append(self.diagnostic_metric_values.get_logging_output())
            else:  # Fallback to the default behavior
                for metric_name, value in self.diagnostic_metric_values.items():
                    formatted_value = self._log_console_format(value, verbose=verbose)
                    diagnostics.append(f"{metric_name}: {formatted_value}")
        return "\n".join(diagnostics)

    @classmethod
    def _log_console_format(cls, n: Number | str | list, verbose: bool = True) -> str:
        """
        Couldn't find nicer & simpler code to format:
        * Full number before the comma,
        * At least 2 significant digits after comma
        * Trunc (not round) after 2 significant digits after comma
        """
        try:
            if n is None:
                return "None"
            if isinstance(n, str):
                return n
            if isinstance(n, list):
                result = ""
                for element in n:
                    result += cls._log_console_format(element, verbose=verbose) + "\n"
                return result
            # For the person that passes a dict to the function. Feel free to determine the output format :)
            if not verbose:
                return cls.__remove_precision(n)
            else:  # We are in verbose
                return str(n)
        except Exception as e:
            logger.error(f"Error formatting output: {e}")
            return str(n)

    @classmethod
    def __remove_precision(cls, n: Number) -> str:  # Can this be removed and replaced by standard python formatting?
        n_str = str(n)
        if "e" in n_str:
            n_str = f"{n:.20f}"
        is_index_after_comma = False
        is_after_first_significant_number = False
        significant_numbers_after_comma = 0
        for index in range(0, len(n_str)):
            c = n_str[index]
            if is_index_after_comma and c != "0":
                is_after_first_significant_number = True
            elif not is_index_after_comma and c == ".":
                is_index_after_comma = True
            if is_after_first_significant_number:
                significant_numbers_after_comma += 1
            if significant_numbers_after_comma > 2:
                return n_str[:index]
        return n_str

    def diagnostics_to_camel_case(self) -> dict[str, Any]:
        """
        Converts diagnostic metric names to camelCase format.
        This is useful for sending diagnostics to Soda Cloud.
        """
        if not self.diagnostic_metric_values:
            return {}

        camel_case_diagnostics = {}
        for key, value in self.diagnostic_metric_values.items():
            camel_case_key = "".join(word.capitalize() if i > 0 else word for i, word in enumerate(key.split("_")))
            camel_case_diagnostics[camel_case_key] = value
        return camel_case_diagnostics


class Measurement:
    def __init__(self, metric_id: str, value: any, metric_name: Optional[str]):
        self.metric_id: str = metric_id
        self.metric_name: Optional[str] = metric_name
        self.value: any = value


class ContractVerificationStatus(Enum):
    UNKNOWN = "UNKNOWN"
    WARNED = "WARNED"
    FAILED = "FAILED"
    PASSED = "PASSED"
    ERROR = "ERROR"


class PostProcessingStageState(Enum):
    ONGOING = "ongoing"
    COMPLETED = "completed"
    FAILED = "failed"


class PostProcessingStage:
    def __init__(self, name: str, stage: PostProcessingStageState):
        self.name: str = name
        self.stage: PostProcessingStageState = stage


class ContractVerificationResult:
    """
    This is the immutable data structure containing all the results from a single contract verification.
    This includes any potential execution errors as well as the results of all the checks performed.

    @param contract: The contract that was verified.
    @param data_source: The data source that was used for the verification.
    @param data_timestamp: The timestamp of the data to use for the verification.
    @param ended_timestamp: The timestamp when the verification ended.
    @param status: The status of the verification. One of ContractVerificationStatus.
    @param measurements: The measurements taken during the verification.
    @param check_results: The results of the checks performed during the verification.
    @param sending_results_to_soda_cloud_failed: If True, sending results to Soda Cloud failed.
    @param log_records: The log records generated during the verification.
    @param post_processing_stages: The post processing stages of the verification.

    """

    def __init__(
        self,
        contract: Contract,
        data_source: DataSource,
        data_timestamp: Optional[datetime],
        started_timestamp: datetime,
        ended_timestamp: datetime,
        status: ContractVerificationStatus,
        measurements: list[Measurement],
        check_results: list[CheckResult],
        sending_results_to_soda_cloud_failed: bool,
        log_records: Optional[list[LogRecord]] = None,
        post_processing_stages: Optional[list[PostProcessingStage]] = None,
    ):
        self.contract: Contract = contract
        self.data_source: DataSource = data_source
        self.data_timestamp: Optional[datetime] = data_timestamp
        self.started_timestamp: datetime = started_timestamp
        self.ended_timestamp: datetime = ended_timestamp
        self.measurements: list[Measurement] = measurements
        self.check_results: list[CheckResult] = check_results
        self.sending_results_to_soda_cloud_failed: bool = sending_results_to_soda_cloud_failed
        self.log_records: Optional[list[LogRecord]] = log_records
        self.status = status
        self.post_processing_stages: Optional[list[PostProcessingStage]] = post_processing_stages

        # Initialze these variables to None, they will be set later when the results are sent to Soda Cloud
        self.scan_id: Optional[str] = None

    def get_logs(self) -> list[str]:
        return [r.getMessage() for r in self.log_records]

    def get_logs_str(self) -> str:
        return "\n".join(self.get_logs())

    def get_errors(self) -> list[str]:
        return [r.getMessage() for r in self.log_records if r.levelno >= ERROR]

    def get_errors_str(self) -> str:
        return "\n".join(self.get_errors())

    def get_warnings(self) -> list[str]:
        return [r.getMessage() for r in self.log_records if r.levelno == WARNING]

    def get_warnings_str(self) -> str:
        return "\n".join(self.get_warnings())

    @property
    def has_errors(self) -> bool:
        return self.status is ContractVerificationStatus.ERROR

    @property
    def is_failed(self) -> bool:
        """
        Returns true if there are checks that have failed.
        False is returned if there are no check results.
        Only looks at check results.
        Ignores execution errors in the logs.
        """
        return self.status is ContractVerificationStatus.FAILED

    @property
    def is_passed(self) -> bool:
        """
        Returns true if there are no checks that have failed.
        Ignores execution errors in the logs.
        """
        return self.status is ContractVerificationStatus.PASSED

    @property
    def is_warned(self) -> bool:
        """
        Returns true if there are checks that have warnings.
        Ignores execution errors in the logs.
        """
        return self.status is ContractVerificationStatus.WARNED

    @property
    def is_ok(self) -> bool:
        return not self.is_failed and not self.has_errors

    @property
    def number_of_checks(self) -> int:
        return len(self.check_results)

    @property
    def number_of_checks_passed(self) -> int:
        return len([check_result for check_result in self.check_results if check_result.outcome == CheckOutcome.PASSED])

    @property
    def number_of_checks_failed(self) -> int:
        return len([check_result for check_result in self.check_results if check_result.outcome == CheckOutcome.FAILED])

    @property
    def number_of_checks_excluded(self) -> int:
        return len(
            [check_result for check_result in self.check_results if check_result.outcome == CheckOutcome.EXCLUDED]
        )



import logging
from typing import Optional

from soda_core.common.logging_constants import soda_logger
from soda_core.common.logs import Logs
from soda_core.common.soda_cloud import SodaCloud
from soda_core.common.yaml import ContractYamlSource, SodaCloudYamlSource
from soda_core.contracts.contract_publication import ContractPublicationResultList
from soda_core.contracts.impl.contract_verification_impl import ContractImpl
from soda_core.contracts.impl.contract_yaml import ContractYaml

logger: logging.Logger = soda_logger


class ContractPublicationImpl:
    def __init__(
        self,
        contract_yaml_sources: list[ContractYamlSource],
        soda_cloud: Optional[SodaCloud],
        soda_cloud_yaml_source: Optional[SodaCloudYamlSource],
        variables: dict[str, str],
        logs: Logs,
    ):
        self.logs: Logs = logs

        self.soda_cloud: Optional[SodaCloud] = soda_cloud
        if self.soda_cloud is None and soda_cloud_yaml_source is not None:
            self.soda_cloud = SodaCloud.from_yaml_source(
                soda_cloud_yaml_source=soda_cloud_yaml_source, provided_variable_values=variables
            )

        self.contract_yamls: list[ContractYaml] = []
        self.contract_impls: list[ContractImpl] = []
        if contract_yaml_sources is None or len(contract_yaml_sources) == 0:
            logger.error(f"No contracts configured")
        else:
            for contract_yaml_source in contract_yaml_sources:
                contract_yaml: ContractYaml = ContractYaml.parse(
                    contract_yaml_source=contract_yaml_source, provided_variable_values=variables
                )
                self.contract_yamls.append(contract_yaml)

    def execute(self) -> ContractPublicationResultList:
        if not self.soda_cloud:
            logger.warning("skipping publication because of missing Soda Cloud configuration")
            return ContractPublicationResultList(items=[], logs=self.logs)
        return ContractPublicationResultList(
            items=[self.soda_cloud.publish_contract(contract_yaml) for contract_yaml in self.contract_yamls],
            logs=self.logs,
        )



from __future__ import annotations

import logging
import threading
from logging import Handler, LogRecord
from typing import Optional

from soda_core.common.logs_base import LogsBase
from soda_core.common.logs_collector import LogsCollector


class Location:
    def __init__(self, file_path: Optional[str], line: Optional[int] = None, column: Optional[int] = None):
        self.file_path: Optional[str] = file_path
        self.line: Optional[int] = line
        self.column: Optional[int] = column

    def __str__(self) -> str:
        src_description: str = self.file_path if self.file_path else "location"
        if self.line is not None and self.column is not None:
            src_description += f"[{self.line},{self.column}]"
        return src_description

    def __hash__(self) -> int:
        return hash((self.line, self.column))

    def get_dict(self) -> dict:
        return {"file_path": self.file_path, "line": self.line, "column": self.column}


class LogCapturer(Handler):
    """
    Captures logging records for the current thread and sends them to the provided LogsBase gatherer.
    """

    def __init__(self, gatherer: LogsBase):
        super().__init__()
        self.gatherer: LogsBase = gatherer
        self.threading_ident: int = threading.get_ident()
        logging.root.addHandler(self)

    def remove_from_root_logger(self) -> None:
        logging.root.removeHandler(self)

    def emit(self, record: LogRecord):
        if self.threading_ident == record.thread:
            self.gatherer.emit(record)


class Logs:
    """
    Configuration manager and entry point for log handling in soda core.
    On creation, it attaches a LogCapturer to the root logger to capture logs for the current thread.
    By default, it uses in-memory LogCollector to gather all logs.
    """

    def __init__(self, gatherer: Optional[LogsBase] = None):
        if gatherer is None:
            gatherer = LogsCollector()
        self.gatherer: LogsBase = gatherer
        self.log_capturer: LogCapturer = LogCapturer(gatherer)

    def remove_from_root_logger(self) -> None:
        self.log_capturer.remove_from_root_logger()

    def __str__(self) -> str:
        return self.get_logs_str()

    def get_log_records(self) -> list[LogRecord]:
        return self.gatherer.get_all_logs()

    def get_logs(self) -> list[str]:
        return [r.getMessage() for r in self.gatherer.get_all_logs()]

    def get_logs_str(self):
        return "\n".join(self.get_logs())

    def get_errors_str(self) -> str:
        return "\n".join(self.get_errors())

    def get_errors(self) -> list[str]:
        return [r.getMessage() for r in self.gatherer.get_error_logs()]

    @property
    def has_errors(self) -> bool:
        return len(self.get_errors()) > 0

    def pop_log_records(self) -> list[LogRecord]:
        log_records: list[LogRecord] = self.get_log_records()
        self.gatherer.close()
        return log_records

    def close(self) -> None:
        self.gatherer.close()



from __future__ import annotations

import numbers
from datetime import date, datetime
from enum import Enum
from typing import Annotated, Any, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Field


class SamplerType(str, Enum):
    ABSOLUTE_LIMIT = "absoluteLimit"


class CheckAttributes(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    check_attributes: list[CheckAttribute] = Field(..., alias="resourceAttributes")


class CheckAttribute(BaseModel):
    name: str
    value: Any

    @classmethod
    def from_raw(cls, name: str, value: Any) -> "CheckAttribute":
        return cls(name=name, value=cls._format_value(value))

    @staticmethod
    def _format_value(value: Any) -> Any:
        # Handle boolean strings explicitly to avoid misinterpretation
        if isinstance(value, str):
            if value.lower() == "true":
                return True
            elif value.lower() == "false":
                return False
        if isinstance(value, bool):
            return value
        if not isinstance(value, datetime) and isinstance(value, date):
            value = datetime.combine(value, datetime.min.time())
        if isinstance(value, datetime):
            if value.tzinfo is None:
                value = value.replace(tzinfo=datetime.now().astimezone().tzinfo)
            return value.isoformat()
        if isinstance(value, numbers.Number):
            return str(value)
        return value


class DatasetConfigurationsDTO(BaseModel):
    model_config = ConfigDict(populate_by_name=True, extra="allow")

    dataset_configurations: list[DatasetConfigurationDTO] = Field(..., alias="datasetConfigurations")


class ComputeWarehouseOverrideDTO(BaseModel):
    model_config = ConfigDict(populate_by_name=True, extra="allow")

    name: str = Field(..., alias="name")


class DatasetConfigurationDTO(BaseModel):
    model_config = ConfigDict(populate_by_name=True, extra="allow")

    collect_failed_rows: Optional[bool] = Field(None, alias="collectFailedRows")
    dataset_qualified_name: Optional[str] = Field(None, alias="datasetQualifiedName")

    # MM and profiling skipped for now, extra=allow will take care of it, this needs to be added as a full structure later
    # metric_monitoring_configuration: Optional[list[dict[str, Any]]] = Field(None, alias="metricMonitoringConfiguration")
    # profiling_configuration: Optional[dict[str, Any]] = Field(None, alias="profilingConfiguration")

    samples_columns: Optional[list[str]] = Field(None, alias="samplesColumns")
    table: Optional[str] = Field(None, alias="table")
    test_row_sampler_configuration: Optional[TestRowSamplerConfigurationDTO] = Field(
        None, alias="testRowSamplerConfiguration"
    )
    compute_warehouse_override: Optional[ComputeWarehouseOverrideDTO] = Field(None, alias="computeWarehouseOverride")


class TestRowSamplerAbsoluteLimitDTO(BaseModel):
    model_config = ConfigDict(populate_by_name=True, extra="allow")

    type: Literal[SamplerType.ABSOLUTE_LIMIT] = Field(..., alias="type")
    limit: int = Field(..., alias="limit")

    def __str__(self):
        return f"AbsoluteLimitRowSampler(limit={self.limit})"


RowSampler = Annotated[
    Union[TestRowSamplerAbsoluteLimitDTO,],
    Field(discriminator="type"),
]


class TestRowSamplerConfigurationDTO(BaseModel):
    model_config = ConfigDict(populate_by_name=True, extra="allow")

    enabled: bool = Field(..., alias="enabled")
    test_row_sampler: Optional[RowSampler] = Field(None, alias="testRowSampler")


class RequestDatasetsConfigurationDTO(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    type: Literal["sodaCoreDatasetsConfiguration"] = Field(
        "sodaCoreDatasetsConfiguration",
        alias="type",
        frozen=True,
    )
    version: Literal[4] = Field(4, alias="version", frozen=True)
    datasets: list[RequestDatasetsConfigurationDatasetDTO] = Field(..., alias="datasets")


class RequestDatasetsConfigurationDatasetDTO(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    data_source: str = Field(..., alias="dataSource")
    dataset_qualified_name: str = Field(..., alias="datasetQualifiedName")
    table: str = Field(..., alias="table")



from __future__ import annotations

import logging
from typing import Optional

from soda_core.common.consistent_hash_builder import ConsistentHashBuilder
from soda_core.common.logging_constants import soda_logger

logger: logging.Logger = soda_logger


class DatasetIdentifier:
    def __init__(self, data_source_name: str, prefixes: list[str], dataset_name: str):
        self.data_source_name = data_source_name
        self.prefixes = prefixes
        self.dataset_name = dataset_name

    @classmethod
    def parse(cls, dataset_qualified_name: Optional[str]) -> DatasetIdentifier:
        from soda_core.common.exceptions import InvalidDatasetQualifiedNameException

        if not dataset_qualified_name:
            raise InvalidDatasetQualifiedNameException("Dataset DQN must be a valid string and cannot be None")

        parts = dataset_qualified_name.split("/")
        if len(parts) < 2:
            raise InvalidDatasetQualifiedNameException("Dataset DQN must contain at least a data source and a dataset")

        data_source_name = parts[0]
        dataset_name = parts[-1]
        prefixes = parts[1:-1] if len(parts) > 2 else []

        return cls(data_source_name, prefixes, dataset_name)

    def to_string(self) -> str:
        return "/".join([self.data_source_name] + self.prefixes + [self.dataset_name])

    def to_hash(self) -> str:
        consistent_hash_builder: ConsistentHashBuilder = ConsistentHashBuilder(hash_string_length=32)  # Use 32 chars
        consistent_hash_builder.add(self.to_string())
        return consistent_hash_builder.get_hash()

    def __repr__(self):
        return (
            f"DatasetIdentifier(data_source='{self.data_source_name}', "
            f"prefixes={self.prefixes}, dataset='{self.dataset_name}')"
        )



from typing import Dict, Optional, Union

from soda_core.common.exceptions import InvalidArgumentException
from soda_core.common.yaml import ContractYamlSource
from soda_core.contracts.contract_verification import (
    ContractVerificationSession,
    ContractVerificationSessionResult,
)
from soda_core.telemetry.soda_telemetry import SodaTelemetry
from typing_extensions import deprecated

soda_telemetry = SodaTelemetry()


@deprecated("Use test_contract_locally instead")
def test_contracts(
    contract_file_paths: Union[str, list[str]],
    variables: Optional[Dict[str, str]] = None,
    verbose: bool = False,
) -> ContractVerificationSessionResult:
    # Limit the contract file paths to only 1. Throw an error if more than 1 is provided.
    if isinstance(contract_file_paths, list) and len(contract_file_paths) > 1:
        raise InvalidArgumentException("Only one contract is allowed at a time")

    return test_contract(
        contract_file_path=contract_file_paths[0],
        variables=variables,
        verbose=verbose,
    )


def test_contract(
    contract_file_path: str = None,
    variables: Optional[Dict[str, str]] = None,
    verbose: bool = False,
) -> ContractVerificationSessionResult:
    contract_yaml_sources = [ContractYamlSource.from_file_path(contract_file_path)]

    contract_verification_result = ContractVerificationSession.execute(
        contract_yaml_sources=contract_yaml_sources,
        variables=variables,
        only_validate_without_execute=True,
        soda_cloud_publish_results=False,
        soda_cloud_use_agent=False,
        soda_cloud_verbose=verbose,
    )

    soda_telemetry.ingest_contract_verification_session_result(
        contract_verification_session_result=contract_verification_result
    )

    return contract_verification_result



from os.path import dirname, exists
from pathlib import Path
from textwrap import dedent
from typing import Optional

from soda_core.cli.exit_codes import ExitCode
from soda_core.common.logging_constants import Emoticons, soda_logger
from soda_core.common.yaml import DataSourceYamlSource


def handle_create_data_source(data_source_file_path: str, data_source_type: str) -> ExitCode:
    soda_logger.info(f"Creating {data_source_type} data source YAML file '{data_source_file_path}'")
    if exists(data_source_file_path):
        soda_logger.error(
            f"Could not create data source file '{data_source_file_path}'. "
            f"File already exists {Emoticons.POLICE_CAR_LIGHT}."
        )
        return ExitCode.LOG_ERRORS
    if data_source_type != "postgres":
        soda_logger.error(f"{Emoticons.POLICE_CAR_LIGHT} Only type postgres is supported atm")
        return ExitCode.LOG_ERRORS
    dir_name = dirname(data_source_file_path)
    try:
        Path(dir_name).mkdir(parents=True, exist_ok=True)
        with open(data_source_file_path, "w") as text_file:
            text_file.write(
                dedent(
                    """
                type: postgres
                name: postgres_ds
                connection:
                    host: localhost
                    user: ${POSTGRES_USERNAME}
                    password: ${POSTGRES_PASSWORD}
                    database: your_postgres_db
                """
                ).strip()
            )
        soda_logger.info(f"{Emoticons.WHITE_CHECK_MARK} Created data source file '{data_source_file_path}'")
        return ExitCode.OK
    except Exception as exc:
        soda_logger.exception(f"An unexpected exception occurred: {exc}")
        return ExitCode.LOG_ERRORS


def handle_test_data_source(data_source_file_path: str) -> ExitCode:
    soda_logger.info(f"Testing data source configuration file {data_source_file_path}")
    from soda_core.common.data_source_impl import DataSourceImpl

    data_source_impl: DataSourceImpl = DataSourceImpl.from_yaml_source(
        DataSourceYamlSource.from_file_path(data_source_file_path)
    )
    error_message: Optional[str] = (
        data_source_impl.test_connection_error_message()
        if data_source_impl
        else "Data source could not be created. See logs above. Or re-run with -v"
    )
    if error_message:
        soda_logger.error(
            f"{Emoticons.POLICE_CAR_LIGHT} Could not connect using data source '{data_source_file_path}': "
            f"{error_message}"
        )
        return ExitCode.LOG_ERRORS
    else:
        soda_logger.info(f"{Emoticons.WHITE_CHECK_MARK} Success! Connection in '{data_source_file_path}' tested ok.")
        return ExitCode.OK



from soda_core.contracts.api.publish_api import publish_contract
from soda_core.contracts.api.test_api import test_contract
from soda_core.contracts.api.verify_api import (
    verify_contract,
    verify_contract_locally,
    verify_contract_on_agent,
    verify_contracts_locally,
    verify_contracts_on_agent,
)

__all__ = [
    "test_contract",
    "verify_contract",
    "verify_contract_locally",
    "verify_contract_on_agent",
    "verify_contracts_locally",
    "verify_contracts_on_agent",
    "publish_contract",
]



from os.path import dirname, exists
from pathlib import Path
from textwrap import dedent
from typing import Optional

from soda_core.cli.exit_codes import ExitCode
from soda_core.common.logging_constants import Emoticons, soda_logger
from soda_core.common.logs import Logs
from soda_core.common.yaml import SodaCloudYamlSource


def handle_create_soda_cloud(soda_cloud_file_path: str) -> ExitCode:
    soda_logger.info(f"Creating Soda Cloud YAML file '{soda_cloud_file_path}'")
    if exists(soda_cloud_file_path):
        soda_logger.error(
            f"Could not create soda cloud file '{soda_cloud_file_path}'. "
            f"File already exists {Emoticons.POLICE_CAR_LIGHT}"
        )
        return ExitCode.LOG_ERRORS
    try:
        Path(dirname(soda_cloud_file_path)).mkdir(parents=True, exist_ok=True)
        with open(soda_cloud_file_path, "w") as text_file:
            text_file.write(
                dedent(
                    """
                soda_cloud:
                  host: cloud.soda.io
                  api_key_id: ${SODA_CLOUD_API_KEY_ID}
                  api_key_secret: ${SODA_CLOUD_API_KEY_SECRET}
                """
                ).strip()
            )
        soda_logger.info(f"{Emoticons.WHITE_CHECK_MARK} Created Soda Cloud configuration file '{soda_cloud_file_path}'")
        return ExitCode.OK
    except Exception as exc:
        soda_logger.exception(f"An unexpected exception occurred: {exc}")
        return ExitCode.LOG_ERRORS


def handle_test_soda_cloud(soda_cloud_file_path: str):
    from soda_core.common.soda_cloud import SodaCloud

    # Start capturing logs
    logs = Logs()

    print(f"Testing soda cloud file {soda_cloud_file_path}")
    soda_cloud_yaml_source: SodaCloudYamlSource = SodaCloudYamlSource.from_file_path(soda_cloud_file_path)
    soda_cloud = SodaCloud.from_yaml_source(soda_cloud_yaml_source, provided_variable_values=None)
    error_msg: Optional[str] = None
    if soda_cloud:
        soda_cloud.test_connection()
        if logs.has_errors:
            error_msg = logs.get_errors_str()
    else:
        error_msg = "Soda Cloud connection could not be created. See logs above. Or re-run with -v"

    if error_msg:
        soda_logger.error(f"{Emoticons.POLICE_CAR_LIGHT} Could not connect to Soda Cloud: {error_msg}")
        return ExitCode.LOG_ERRORS
    else:
        soda_logger.info(
            f"{Emoticons.WHITE_CHECK_MARK} Success! Tested Soda Cloud credentials in '{soda_cloud_file_path}'"
        )
        return ExitCode.OK



import re
from datetime import datetime, timezone
from typing import Optional


def convert_datetime_to_str(dt: datetime) -> Optional[str]:
    try:
        new_dt = interpret_datetime_as_utc(dt)
        return new_dt.isoformat(timespec="seconds")

        # OLD CODE: to remove
        # Attention, if dt does not have timezone info, Python will assume dt is in the local system timezone (e.g. UTC+1)
        # return dt.astimezone(timezone.utc).isoformat(timespec="seconds")
    except Exception as e:
        return None


def interpret_datetime_as_utc(dt: datetime) -> datetime:
    # First, we check if the datetime has timezone info
    if dt.tzinfo is None:
        # If dt does not have timezone info, Python will assume dt is in the local system timezone (e.g. UTC+1)
        # Then we need to REPLACE (not astimezone) the timezone to UTC
        new_dt = dt.replace(tzinfo=timezone.utc)
    else:
        # The datetime has timezone info, we need to **convert** it to UTC (to be consistent with the rest of the code)
        new_dt = dt.astimezone(timezone.utc)
    return new_dt


def convert_str_to_datetime(date_string: str) -> Optional[datetime]:
    """
    Returns the parsed datetime or None if there is a parsing error.
    No exceptions are raised
    """
    try:
        # fromisoformat raises ValueError if date_string ends with a Z:
        # Eg Invalid isoformat string: '2025-02-21T06:16:59Z'
        if date_string.endswith("Z"):
            # Z means Zulu time, which is UTC
            # Converting timezone to format that fromisoformat understands
            datetime_str_without_z: str = date_string[:-1]
            datetime_str_without_z_seconds = re.sub(r"\.(\d+)$", "", datetime_str_without_z)
            date_string = f"{datetime_str_without_z_seconds}+00:00"
        return datetime.fromisoformat(date_string)
    except (Exception, AttributeError) as e:
        return None



from __future__ import annotations

import logging
from textwrap import dedent
from typing import Optional

from soda_core.common.logging_constants import ExtraKeys, soda_logger
from soda_core.common.yaml import YamlObject
from soda_core.contracts.impl.contract_yaml import (
    CheckYaml,
    CheckYamlParser,
    ColumnYaml,
    MissingAncValidityCheckYaml,
)

logger: logging.Logger = soda_logger


class MetricCheckYamlParser(CheckYamlParser):
    def get_check_type_names(self) -> list[str]:
        return ["metric"]

    def parse_check_yaml(
        self, check_type_name: str, check_yaml_object: YamlObject, column_yaml: Optional[ColumnYaml]
    ) -> Optional[CheckYaml]:
        return MetricCheckYaml(type_name=check_type_name, check_yaml_object=check_yaml_object)


class MetricCheckYaml(MissingAncValidityCheckYaml):
    def __init__(self, type_name: str, check_yaml_object: YamlObject):
        super().__init__(type_name=type_name, check_yaml_object=check_yaml_object)
        self.expression: Optional[str] = check_yaml_object.read_string_opt("expression")
        if self.expression:
            self.expression = dedent(self.expression).strip()
        self.query: Optional[str] = check_yaml_object.read_string_opt("query")
        if self.query:
            self.query = dedent(self.query).strip()
        if not self.expression and not self.query:
            logger.error(
                msg="In a 'metric' check, either 'expression' or 'query' is required",
                extra={ExtraKeys.LOCATION: check_yaml_object.location},
            )



from __future__ import annotations

import copy
import logging
import os
import re
from enum import Enum
from io import StringIO
from numbers import Number
from typing import Iterable, Optional

from ruamel.yaml import YAML, CommentedMap, CommentedSeq
from ruamel.yaml.error import MarkedYAMLError
from soda_core.common.exceptions import YamlParserException
from soda_core.common.logging_constants import ExtraKeys, soda_logger
from soda_core.common.logs import Location

logger: logging.Logger = soda_logger


class YamlParser:
    def __init__(self):
        self.ruamel_yaml_parser: YAML = YAML()
        self.ruamel_yaml_parser.preserve_quotes = True

    def dump_to_string(self, data):
        stream = StringIO()
        self.ruamel_yaml_parser.dump(data, stream)
        return stream.getvalue()


class FileType(str, Enum):
    DATA_SOURCE = "Data Source"
    SODA_CLOUD = "Soda Cloud"
    CONTRACT = "Contract"
    YAML = "YAML"


class YamlSource:
    """
    YamlSource is an abstraction for multiple types of YAML sources: see subclasses below.
    It does lazy loading of the file so that errors are only generated when resolving or parsing.
    It's a stateful object that allows for optional variable resolving.

    Usage:


    yaml_source: YamlSource = ContractYamlSource.from_str(
        yaml_str="...yaml...",
        file_path="./some/contract.yml" # optional
    )
    # or
    yaml_source: YamlSource = ContractYamlSource.from_file_path(
        file_path="./some/contract.yml"
    )

    # The default file type is 'YAML'.
    # Specific subclasses use different file types.

    # Optionally resolve variables as many times as you like:
    # resolving will trigger file loading (if applicable)
    yaml_source.resolve(variables={}, use_env_vars=True)

    # parse will create the YamlObject
    # resolving will trigger file loading (if applicable)
    yaml_object: YamlObject = yaml_source.parse()
    """

    __yaml_parser = YamlParser()
    __file_type = FileType.YAML.value

    @classmethod
    def from_str(cls, yaml_str: str, file_path: Optional[str] = None) -> YamlSource:
        """
        Raises an assertion exception if yaml_str is not a str
        """
        assert isinstance(yaml_str, str)
        return cls(file_path=file_path, yaml_str=yaml_str)

    @classmethod
    def from_file_path(cls, file_path: str) -> YamlSource:
        """
        Raises an assertion exception if file_path is not a str
        """
        assert isinstance(file_path, str)
        return cls(file_path=file_path)

    def __init__(self, file_path: Optional[str] = None, yaml_str: Optional[str] = None):
        self.file_type: str = self.__file_type
        self.file_path: str = file_path
        self.description: str = self._build_description(self.file_type, self.file_path)
        self.is_file_read: bool = False
        self.yaml_str: Optional[str] = yaml_str
        self.yaml_str_original: Optional[str] = yaml_str
        self.resolve_on_read: bool = False
        self.resolve_on_read_variable_values: Optional[dict[str, str]] = None
        self.resolve_on_read_soda_variable_values: Optional[dict[str, str]] = None
        self.resolve_on_read_use_env_vars: bool = True

    def __init_subclass__(cls, file_type: FileType, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.__file_type = file_type.value

    def __str__(self) -> str:
        return self.description

    @classmethod
    def _build_description(cls, file_type: str, file_path: Optional[str]) -> str:
        if file_type:
            if file_path:
                return f"{file_type} file '{file_path}'"
            else:
                return f"{file_type} YAML string"
        else:
            if file_path:
                return f"YAML file '{file_path}'"
            else:
                return f"YAML string"

    def _ensure_yaml_str(self) -> None:
        if not isinstance(self.yaml_str, str) and isinstance(self.file_path, str) and not self.is_file_read:
            # There should be only 1 reading attempt.
            self.is_file_read = True
            try:
                with open(self.file_path) as f:
                    self.yaml_str = f.read()
                    self.yaml_str_original = self.yaml_str
                    if len(self.yaml_str) == 0:
                        logger.error(f"{self.description} is empty")

            except OSError as e:
                if not os.path.exists(self.file_path):
                    logger.error(f"{self.description} does not exist")
                elif not os.path.isdir(self.file_path):
                    logger.error(f"{self.description} is a directory")
                else:
                    logger.error(f"{self.description} can't be read", exc_info=True)

    def resolve(self, variables: Optional[dict[str, str]] = None, use_env_vars: bool = True) -> None:
        """
        Note: this does not change the object, but returns a new ResolvedYamlSource instead
        """
        self._ensure_yaml_str()
        self.yaml_str = VariableResolver.resolve(
            source_text=self.yaml_str, variable_values=variables, use_env_vars=use_env_vars
        )

    def resolve_on_read_value(
        self,
        resolved_variable_values: Optional[dict[str, str]],
        soda_values: Optional[dict[str, str]],
        use_env_vars: bool,
    ):
        self.resolve_on_read = True
        self.resolve_on_read_variable_values = resolved_variable_values
        self.resolve_on_read_soda_variable_values = soda_values
        self.resolve_on_read_use_env_vars = use_env_vars

    def parse(self) -> YamlObject:
        self._ensure_yaml_str()
        location = Location(file_path=self.file_path)
        if not isinstance(self.yaml_str, str):
            raise YamlParserException("YAML source is not a string", str(location))

        try:
            root_yaml_object: any = self.__yaml_parser.ruamel_yaml_parser.load(self.yaml_str)
            if isinstance(root_yaml_object, dict):
                return YamlObject(self, root_yaml_object)
            else:
                yaml_type: str = root_yaml_object.__class__.__name__ if root_yaml_object is not None else "empty"
                yaml_type = "a list" if yaml_type == "CommentedSeq" else yaml_type
                raise YamlParserException(
                    f"{self.description} root must be an object, " f"but was {yaml_type}", str(location)
                )

        except MarkedYAMLError as e:
            mark = e.context_mark if e.context_mark else e.problem_mark
            line = mark.line + 1
            col = mark.column + 1
            location = Location(file_path=self.file_path, line=line, column=col)
            raise YamlParserException(f"YAML syntax error", str(location))


class DataSourceYamlSource(YamlSource, file_type=FileType.DATA_SOURCE):
    ...


class SodaCloudYamlSource(YamlSource, file_type=FileType.SODA_CLOUD):
    ...


class ContractYamlSource(YamlSource, file_type=FileType.CONTRACT):
    ...


class YamlValue:
    def __init__(self, yaml_source: YamlSource) -> None:
        self.yaml_source: YamlSource = yaml_source

    def _yaml_wrap(self, value: any, location: Optional[Location] = None):
        # Resolve variables on read if configured to do so.
        # Only resolve one level deep in dicts and lists, we are not building a full template engine here.

        if value is None:
            return None

        # To avoid modifying the original value during variable resolution
        wrapped_value = copy.deepcopy(value) if isinstance(value, (dict, list)) else value

        if isinstance(wrapped_value, dict):
            if self.yaml_source.resolve_on_read:
                for k, v in wrapped_value.items():
                    if isinstance(v, str):
                        wrapped_value[k] = self._resolve_variable(
                            source_text=v,
                            location=location,
                        )
            return YamlObject(yaml_source=self.yaml_source, yaml_dict=wrapped_value)
        elif isinstance(wrapped_value, list):
            if self.yaml_source.resolve_on_read:
                for i in range(0, len(wrapped_value)):
                    v = wrapped_value[i]
                    if isinstance(v, str):
                        wrapped_value[i] = self._resolve_variable(
                            source_text=v,
                            location=location,
                        )
            return YamlList(yaml_source=self.yaml_source, yaml_list=wrapped_value)
        elif isinstance(wrapped_value, str) and self.yaml_source.resolve_on_read:
            wrapped_value = self._resolve_variable(
                source_text=wrapped_value,
                location=location,
            )
        return wrapped_value

    def _resolve_variable(
        self,
        source_text: str,
        location: Optional[Location] = None,
    ) -> str:
        return VariableResolver.resolve(
            source_text=source_text,
            variable_values=self.yaml_source.resolve_on_read_variable_values,
            soda_variable_values=self.yaml_source.resolve_on_read_soda_variable_values,
            use_env_vars=self.yaml_source.resolve_on_read_use_env_vars,
            location=location,
        )

    @classmethod
    def yaml_unwrap(cls, o) -> Optional[object]:
        if isinstance(o, YamlObject):
            return {k: cls.yaml_unwrap(v) for k, v in o.items()}
        if isinstance(o, YamlList):
            return [cls.yaml_unwrap(e) for e in o]
        return o


def get_location(yaml_value: any, yaml_file_path: Optional[str]) -> Optional[Location]:
    if isinstance(yaml_value, CommentedMap) or isinstance(yaml_value, CommentedSeq):
        return Location(file_path=yaml_file_path, line=yaml_value.lc.line, column=yaml_value.lc.col)


class YamlObject(YamlValue):
    def __init__(self, yaml_source: YamlSource, yaml_dict: dict) -> None:
        super().__init__(yaml_source)
        self.yaml_dict: dict = yaml_dict
        self.location: Optional[Location] = get_location(self.yaml_dict, yaml_source.file_path)

    def items(self) -> list[tuple]:
        return [(k, self._yaml_wrap(v)) for k, v in self.yaml_dict.items()]

    def keys(self) -> list[str]:
        return list(self.yaml_dict.keys())

    def __iter__(self) -> iter:
        return iter(self.yaml_dict.keys())

    def has_key(self, key: str) -> bool:
        return key in self.yaml_dict

    # TODO: should we move this into the `ContractYaml` class?
    def read_dataset_identifier(self, yaml_key: str, required: bool = True) -> Optional[str]:
        dataset_qualified_name: Optional[str] = self.read_string_opt(yaml_key)
        if isinstance(dataset_qualified_name, str):
            parts = dataset_qualified_name.split("/")
            if len(parts) < 2:
                logger.error(
                    msg=f"Invalid dataset qualified name in '{yaml_key}': '{dataset_qualified_name}' "
                    f"must be slash-separated, fully qualified dataset name.",
                    extra={ExtraKeys.LOCATION: self.create_location_from_yaml_dict_key(yaml_key)},
                )
                return None
        elif required:
            raise YamlParserException("The YAML is missing the required 'dataset' property", str(self.location))
        return dataset_qualified_name

    def read_object(self, key: str) -> Optional[YamlObject]:
        """
        An error is generated if the value is missing or not a YAML object.
        :return: a YamlObject if the value for the key is a YAML object, otherwise None.
        """
        return self.read_value(key=key, expected_type=dict, required=True, default_value=None)

    def read_object_opt(self, key: str, default_value: Optional[dict] = None) -> Optional[YamlObject]:
        """
        An error is generated if the value is present and not a YAML object.
        :return: a YamlObject if the value for the key is a YAML object, otherwise None.
        """
        return self.read_value(key=key, expected_type=dict, required=False, default_value=default_value)

    def read_list(
        self, key: str, expected_element_type: Optional[type] = None, required: bool = True
    ) -> Optional[YamlList]:
        """
        An error is generated if the value is missing or not a YAML list.
        :return: a list if the value for the key is a YAML list, otherwise None.
        """
        list_value: Optional[YamlList] = self.read_value(
            key=key, expected_type=list, required=required, default_value=None
        )
        if isinstance(list_value, YamlList):
            filtered_list: list = []
            if isinstance(expected_element_type, type):
                for index, element in enumerate(iter(list_value)):
                    if isinstance(element, expected_element_type):
                        filtered_list.append(element)
                    else:
                        logger.error(
                            msg=(
                                f"YAML key '{key}' expects a "
                                f"list of {expected_element_type.__name__}. "
                                f"But element {index} was {type(element).__name__}"
                            ),
                            extra={
                                ExtraKeys.LOCATION: self.create_location_from_yaml_dict_key(key),
                            },
                        )
            list_value.yaml_list = filtered_list
        return list_value

    def read_list_opt(self, key: str) -> Optional[YamlList]:
        """
        An error is generated if the value is present and not a YAML list.
        :return: a list if the value for the key is a YAML list, otherwise None.
        """
        return self.read_value(key=key, expected_type=list, required=False, default_value=None)

    def read_list_of_objects_opt(self, key: str) -> Optional[YamlList]:
        return self.read_list_of_objects(key) if key in self.yaml_dict else None

    def read_list_of_objects(self, key: str) -> Optional[YamlList]:
        list_value: Optional[YamlList] = self.read_list(key=key, expected_element_type=YamlObject)
        if isinstance(list_value, YamlList):
            for i in range(0, len(list_value.yaml_list)):
                element = list_value.yaml_list[i]
                if not isinstance(element, YamlObject):
                    logger.error(
                        msg=(
                            f"YAML key '{key}' expected list of objects.  "
                            f"But element {i} was {type(element).__name__}"
                        ),
                        extra={
                            ExtraKeys.LOCATION: self.create_location_from_yaml_dict_key(key),
                        },
                    )
                    list_value.yaml_list[i] = None
        return list_value

    def read_list_of_strings(self, key: str) -> list[str] | None:
        l = self.read_list(key, expected_element_type=str)
        return l.to_list() if isinstance(l, YamlList) else None

    def read_list_of_strings_opt(self, key: str) -> list[str] | None:
        l = self.read_list(key, expected_element_type=str, required=False)
        return l.to_list() if isinstance(l, YamlList) else None

    def read_string(self, key: str, env_var: Optional[str] = None, required: bool = True) -> Optional[str]:
        """
        An error is generated if the value is missing or not a string.
        :return: a str if the value for the key is a YAML string, otherwise None.
        """
        return self.read_value(key=key, env_var=env_var, expected_type=str, required=required, default_value=None)

    def read_string_opt(
        self, key: str, env_var: Optional[str] = None, default_value: Optional[str] = None
    ) -> Optional[str]:
        """
        An error is generated if the value is present and not a string.
        :return: a str if the value for the key is a string, otherwise None.
        """
        return self.read_value(key=key, env_var=env_var, expected_type=str, required=False, default_value=default_value)

    def read_bool(self, key: str) -> Optional[bool]:
        """
        An appropriate error log is generated if the value is not a bool or if the key is missing
        :return: a bool if the value for the key is a YAML boolean, otherwise None.
        """
        return self.read_value(key=key, expected_type=bool, required=True, default_value=None)

    def read_bool_opt(self, key: str, default_value: Optional[bool] = None) -> Optional[bool]:
        """
        An appropriate error log is generated if the value is not a bool.
        :return: a bool if the value for the key is a YAML boolean, otherwise None.
        """
        return self.read_value(key=key, expected_type=bool, required=False, default_value=default_value)

    def read_number(self, key: str) -> Optional[Number]:
        """
        An appropriate error log is generated if the value is not a number or if the key is missing
        :return: a bool if the value for the key is a YAML number, otherwise None.
        """
        return self.read_value(key=key, expected_type=Number, required=True, default_value=None)

    def read_number_opt(self, key: str, default_value: Optional[Number] = None) -> Optional[Number]:
        """
        An appropriate error log is generated if the value is not a number.
        :return: a Number if the value for the key is a YAML number, otherwise None.
        """
        return self.read_value(key=key, expected_type=Number, required=False, default_value=default_value)

    def read_value(
        self,
        key: str,
        env_var: Optional[str] = None,
        expected_type: type | list[type] = None,
        required: bool = False,
        default_value=None,
    ) -> Optional[object]:
        location: Location = self.create_location_from_yaml_dict_key(key) if key in self.yaml_dict else self.location

        key_description: str = f"YAML key '{key}'"
        if env_var is not None and env_var in os.environ:
            key_description = f"Env var '{env_var}'"
            if key in self.yaml_dict:
                logger.debug(f"Using environment var '{env_var}' instead of configuration key '{key}'")
            value = os.environ[env_var]
        elif key in self.yaml_dict:
            value = self.yaml_dict.get(key)
        else:
            if required:
                raise YamlParserException(f"{key_description} is required", str(location))
            value = default_value

        expected_types: list[type] = []
        if expected_type and not isinstance(expected_type, list):
            expected_types = [expected_type]
        elif isinstance(expected_type, list):
            expected_types = expected_type

        if expected_types and not any(isinstance(value, t) for t in expected_types) and value != default_value:
            actual_type_str: str = type(value).__name__
            if value is None:
                actual_type_str = "None"
            elif isinstance(value, dict):
                actual_type_str = "YAML object"
            elif isinstance(value, list):
                actual_type_str = "YAML list"
            logger.error(
                msg=f"{key_description} expected one of {[et.__name__ for et in expected_types]}, "
                f"but was {actual_type_str}",
                extra={
                    ExtraKeys.LOCATION: location,
                },
            )
            value = None

        return self._yaml_wrap(value, location=location)

    def create_location_from_yaml_dict_key(self, key) -> Optional[Location]:
        if isinstance(self.yaml_dict, CommentedMap):
            if key in self.yaml_dict:
                ruamel_location = self.yaml_dict.lc.value(key)
                line: int = ruamel_location[0]
                column: int = ruamel_location[1]
                return Location(file_path=self.yaml_source.file_path, line=line, column=column)

    def to_dict(self) -> dict:
        return self.yaml_dict


class YamlList(YamlValue, Iterable):
    def __init__(self, yaml_source: YamlSource, yaml_list: list) -> None:
        super().__init__(yaml_source)
        self.yaml_list: list = yaml_list
        self.location: Optional[Location] = get_location(yaml_list, yaml_source.file_path)

    def __iter__(self) -> iter:
        return iter([self._yaml_wrap(element) for element in self.yaml_list])

    def to_list(self) -> list:
        return self.yaml_list

    def create_location_from_yaml_list_index(self, index) -> Optional[Location]:
        if index < len(self.yaml_list):
            ruamel_location = self.yaml_list.lc.item(index)
            line: int = ruamel_location[0]
            column: int = ruamel_location[1]
            return Location(file_path=self.yaml_source.file_path, line=line, column=column)


class VariableResolver:
    @classmethod
    def resolve(
        cls,
        source_text: str,
        variable_values: Optional[dict[str, str]] = None,
        soda_variable_values: Optional[dict[str, str]] = None,
        use_env_vars: bool = True,
        location: Optional[Location] = None,
    ) -> str:
        if isinstance(source_text, str):
            # First pass: sometimes the value is just the variable with quotes. If so, we can just return the value directly, no casting to string needed.
            pattern = r"\$\{ *([a-z]+)\.([a-zA-Z_][a-zA-Z_0-9]*) *\}"
            match = re.fullmatch(pattern, source_text)
            if match:
                return cls.get_variable(
                    namespace=match.group(1).strip(),
                    variable=match.group(2).strip(),
                    variable_values=variable_values,
                    soda_variable_values=soda_variable_values,
                    use_env_vars=use_env_vars,
                    location=location,
                )

            # Otherwise do a regex replace for all variable patterns in the string, which may be mixed with other text.
            return re.sub(
                pattern=pattern,
                repl=lambda m: cls._resolve_variable_pattern(
                    namespace=m.group(1).strip(),
                    variable=m.group(2).strip(),
                    variable_values=variable_values,
                    soda_variable_values=soda_variable_values,
                    use_env_vars=use_env_vars,
                    location=location,
                ),
                string=source_text,
            )
        else:
            return source_text

    @classmethod
    def _resolve_variable_pattern(
        cls,
        namespace: str,
        variable: str,
        variable_values: Optional[dict[str, str]],
        soda_variable_values: Optional[dict[str, str]],
        use_env_vars: bool,
        location: Optional[Location] = None,
    ) -> str:
        value: Optional[str] = cls.get_variable(
            namespace=namespace,
            variable=variable,
            variable_values=variable_values,
            soda_variable_values=soda_variable_values,
            use_env_vars=use_env_vars,
            location=location,
        )
        # If the variable is not found, return the original expression. Cast to string in case the value is a number so that we can re-insert it into the text.
        return str(value) if isinstance(value, str) or isinstance(value, Number) else f"${{{namespace}.{variable}}}"

    @classmethod
    def get_variable(
        cls,
        namespace: str,
        variable: str,
        variable_values: Optional[dict[str, str]],
        soda_variable_values: Optional[dict[str, str]],
        use_env_vars: bool = True,
        location: Optional[Location] = None,
    ) -> Optional[str]:
        if namespace == "var":
            if not isinstance(variable_values, dict) or variable not in variable_values:
                logger.error(
                    msg=(f"Variable '{variable}' was used and not declared"),
                    extra={ExtraKeys.LOCATION: location} if location else None,
                )
            else:
                return variable_values[variable]
        elif namespace == "soda":
            if not isinstance(soda_variable_values, dict) or variable not in soda_variable_values:
                logger.error(
                    msg=(f"Variable '{variable}' was used and not available in the 'soda' namespace"),
                    extra={ExtraKeys.LOCATION: location} if location else None,
                )
            else:
                return soda_variable_values[variable]
        elif namespace == "env":
            if use_env_vars:
                return os.getenv(variable) if variable in os.environ else None
            else:
                logger.error(
                    msg=(
                        f"Environment variable '{variable}' will not be resolved because environment variables are "
                        f"not supported inside contract."
                    ),
                    extra={ExtraKeys.LOCATION: location} if location else None,
                )
        return None


def yaml_to_string(yaml_value: dict) -> str:
    text_stream = StringIO()
    yaml = YAML()
    yaml.indent(mapping=2, sequence=4, offset=2)
    yaml.dump(yaml_value, text_stream)
    text_stream.seek(0)

    return text_stream.read()



from __future__ import annotations

import logging

from soda_core.common.data_source_impl import DataSourceImpl
from soda_core.common.logging_constants import soda_logger
from soda_core.common.sql_dialect import *
from soda_core.contracts.contract_verification import CheckOutcome, CheckResult
from soda_core.contracts.impl.check_types.duplicate_check_yaml import (
    ColumnDuplicateCheckYaml,
    MultiColumnDuplicateCheckYaml,
)
from soda_core.contracts.impl.check_types.invalidity_check_yaml import InvalidCheckYaml
from soda_core.contracts.impl.check_types.missing_check import MissingCountMetricImpl
from soda_core.contracts.impl.check_types.row_count_check import RowCountMetricImpl
from soda_core.contracts.impl.contract_verification_impl import (
    AggregationMetricImpl,
    CheckImpl,
    CheckParser,
    ColumnImpl,
    ContractImpl,
    DerivedMetricImpl,
    DerivedPercentageMetricImpl,
    MeasurementValues,
    MetricImpl,
    MissingAndValidity,
    MissingAndValidityCheckImpl,
    ThresholdImpl,
    ThresholdType,
)

logger: logging.Logger = soda_logger


class DuplicateCheckParser(CheckParser):
    def get_check_type_names(self) -> list[str]:
        return ["duplicate"]

    def parse_check(
        self,
        contract_impl: ContractImpl,
        column_impl: Optional[ColumnImpl],
        check_yaml: InvalidCheckYaml,
    ) -> Optional[CheckImpl]:
        if column_impl:
            return ColumnDuplicateCheckImpl(
                contract_impl=contract_impl,
                column_impl=column_impl,
                check_yaml=check_yaml,
            )
        else:
            return MultiColumnDuplicateCheckImpl(
                contract_impl=contract_impl,
                check_yaml=check_yaml,
            )


class ColumnDuplicateCheckImpl(MissingAndValidityCheckImpl):
    def __init__(
        self,
        contract_impl: ContractImpl,
        column_impl: ColumnImpl,
        check_yaml: ColumnDuplicateCheckYaml,
    ):
        super().__init__(
            contract_impl=contract_impl,
            column_impl=column_impl,
            check_yaml=check_yaml,
        )
        self.threshold = ThresholdImpl.create(
            threshold_yaml=check_yaml.threshold,
            default_threshold=ThresholdImpl(type=ThresholdType.SINGLE_COMPARATOR, must_be=0),
        )
        self.metric_name = "duplicate_percent" if check_yaml.metric == "percent" else "duplicate_count"

    def setup_metrics(
        self,
        contract_impl: ContractImpl,
        column_impl: ColumnImpl,
        check_yaml: ColumnDuplicateCheckYaml,
    ):
        self.distinct_count_metric_impl: MetricImpl = self._resolve_metric(
            ColumnDistinctCountMetricImpl(contract_impl=contract_impl, column_impl=column_impl, check_impl=self)
        )

        self.check_rows_tested_metric_impl = self._resolve_metric(
            RowCountMetricImpl(
                contract_impl=contract_impl,
                check_impl=self,
            )
        )

        self.missing_count_metric_impl = self._resolve_metric(
            MissingCountMetricImpl(contract_impl=contract_impl, column_impl=column_impl, check_impl=self)
        )

        self.duplicate_count_metric_impl = self._resolve_metric(
            DuplicateCountMetricImpl(
                metric_type="duplicate_count",
                distinct_count_metric_impl=self.distinct_count_metric_impl,
                valid_count_metric_impl=self.check_rows_tested_metric_impl,
                check_filter=self.check_yaml.filter,
                missing_and_validity=self.missing_and_validity,
            )
        )

        self.duplicate_percent_metric_impl = self._resolve_metric(
            DerivedPercentageMetricImpl(
                metric_type="duplicate_percent",
                fraction_metric_impl=self.duplicate_count_metric_impl,
                total_metric_impl=self.check_rows_tested_metric_impl,
            )
        )

    def evaluate(self, measurement_values: MeasurementValues) -> CheckResult:
        outcome: CheckOutcome = CheckOutcome.NOT_EVALUATED

        diagnostic_metric_values: dict[str, float] = {"dataset_rows_tested": self.contract_impl.dataset_rows_tested}

        distinct_count: int = measurement_values.get_value(self.distinct_count_metric_impl)

        check_rows_tested_count: int = measurement_values.get_value(self.check_rows_tested_metric_impl)
        if isinstance(check_rows_tested_count, Number):
            diagnostic_metric_values["check_rows_tested"] = check_rows_tested_count

        missing_count: int = measurement_values.get_value(self.missing_count_metric_impl)
        if isinstance(missing_count, Number):
            diagnostic_metric_values["missing_count"] = missing_count

        duplicate_count: int = 0
        duplicate_percent: float = 0
        if (
            isinstance(check_rows_tested_count, Number)
            and isinstance(missing_count, Number)
            and isinstance(distinct_count, Number)
        ):
            duplicate_count: int = check_rows_tested_count - missing_count - distinct_count
            diagnostic_metric_values["duplicate_count"] = duplicate_count

            non_missing_total: int = check_rows_tested_count - missing_count
            if non_missing_total > 0:
                duplicate_percent: float = duplicate_count * 100 / non_missing_total
            diagnostic_metric_values["duplicate_percent"] = duplicate_percent

        threshold_value: Optional[Number] = (
            duplicate_percent if self.metric_name == "duplicate_percent" else duplicate_count
        )

        outcome = self.evaluate_threshold(threshold_value)

        return CheckResult(
            check=self._build_check_info(),
            outcome=outcome,
            threshold_value=threshold_value,
            diagnostic_metric_values=diagnostic_metric_values,
        )


class ColumnDistinctCountMetricImpl(AggregationMetricImpl):
    def __init__(
        self,
        contract_impl: ContractImpl,
        column_impl: ColumnImpl,
        check_impl: MissingAndValidityCheckImpl,
    ):
        super().__init__(
            contract_impl=contract_impl,
            column_impl=column_impl,
            metric_type="distinct_count",
            check_filter=check_impl.check_yaml.filter,
            missing_and_validity=check_impl.missing_and_validity,
        )

    def sql_expression(self) -> SqlExpression:
        column_name: str = self.column_impl.column_yaml.name
        filters: list[SqlExpression] = [SqlExpressionStr.optional(self.check_filter)]
        if self.missing_and_validity:
            filters.append(
                NOT(
                    OR.optional(
                        [
                            self.missing_and_validity.is_missing_expr(column_name),
                        ]
                    )
                )
            )
        filter_expr: Optional[SqlExpression] = AND.optional(filters)
        if filter_expr:
            return COUNT(DISTINCT(CASE_WHEN(filter_expr, column_name)))
        else:
            return COUNT(DISTINCT(column_name))

    def convert_db_value(self, value) -> int:
        # Note: expression SUM(CASE WHEN "id" IS NULL THEN 1 ELSE 0 END) gives NULL / None as a result if
        # there are no rows
        return int(value) if value is not None else 0


class DuplicateCountMetricImpl(DerivedMetricImpl):
    def __init__(
        self,
        metric_type: str,
        distinct_count_metric_impl: MetricImpl,
        valid_count_metric_impl: MetricImpl,
        check_filter: Optional[str],
        missing_and_validity: Optional[MissingAndValidity],
    ):
        self.distinct_count_metric_impl: MetricImpl = distinct_count_metric_impl
        self.valid_count_metric_impl: MetricImpl = valid_count_metric_impl
        # Mind the ordering as the self._build_id() must come last
        super().__init__(
            contract_impl=distinct_count_metric_impl.contract_impl,
            column_impl=distinct_count_metric_impl.column_impl,
            metric_type=metric_type,
            check_filter=check_filter,
            missing_and_validity=missing_and_validity,
        )

    def get_metric_dependencies(self) -> list[MetricImpl]:
        return [self.distinct_count_metric_impl, self.valid_count_metric_impl]

    def compute_derived_value(self, measurement_values: MeasurementValues) -> Number:
        distinct_count: int = measurement_values.get_value(self.distinct_count_metric_impl)
        valid_count: int = measurement_values.get_value(self.valid_count_metric_impl)
        return valid_count - distinct_count


class MultiColumnDuplicateCheckImpl(CheckImpl):
    def __init__(
        self,
        contract_impl: ContractImpl,
        check_yaml: MultiColumnDuplicateCheckYaml,
    ):
        super().__init__(
            contract_impl=contract_impl,
            column_impl=None,
            check_yaml=check_yaml,
        )
        self.threshold = ThresholdImpl.create(
            threshold_yaml=check_yaml.threshold,
            default_threshold=ThresholdImpl(type=ThresholdType.SINGLE_COMPARATOR, must_be=0),
        )

    def setup_metrics(
        self,
        contract_impl: ContractImpl,
        column_impl: Optional[ColumnImpl],
        check_yaml: MultiColumnDuplicateCheckYaml,
    ):
        self.metric_name = "duplicate_percent" if check_yaml.metric == "percent" else "duplicate_count"

        self.multi_column_distinct_count_metric_impl: MetricImpl = self._resolve_metric(
            MultiColumnDistinctCountMetricImpl(
                contract_impl=contract_impl, check_impl=self, column_names=check_yaml.columns
            )
        )

        self.row_count_metric_impl = self._resolve_metric(
            RowCountMetricImpl(
                contract_impl=contract_impl,
                check_impl=self,
            )
        )

        # self.duplicate_count_metric_impl = self._resolve_metric(
        #     MultiColumnDuplicateCountMetricImpl(
        #         metric_type="duplicate_count",
        #         multi_column_distinct_count_metric_impl=self.multi_column_distinct_count_metric_impl,
        #         row_count_metric_impl=self.row_count_metric_impl,
        #         check_filter=self.check_yaml.filter,
        #     )
        # )
        #
        # self.duplicate_percent_metric_impl = self._resolve_metric(
        #     DerivedPercentageMetricImpl(
        #         metric_type="duplicate_percent",
        #         fraction_metric_impl=self.duplicate_count_metric_impl,
        #         total_metric_impl=self.row_count_metric_impl,
        #     )
        # )

    def evaluate(self, measurement_values: MeasurementValues) -> CheckResult:
        outcome: CheckOutcome = CheckOutcome.NOT_EVALUATED

        row_count: int = measurement_values.get_value(self.row_count_metric_impl)
        distinct_count: int = measurement_values.get_value(self.multi_column_distinct_count_metric_impl)
        duplicate_count: int = 0
        duplicate_percent: float = 0

        if isinstance(row_count, Number) and isinstance(distinct_count, Number):
            duplicate_count = row_count - distinct_count
            if row_count > 0:
                duplicate_percent = duplicate_count * 100 / row_count

        diagnostic_metric_values: dict[str, float] = {
            "duplicate_count": duplicate_count,
            "duplicate_percent": duplicate_percent,
            "check_rows_tested": row_count,
            "dataset_rows_tested": self.contract_impl.dataset_rows_tested,
        }

        threshold_value: Optional[Number] = (
            duplicate_percent if self.metric_name == "duplicate_percent" else duplicate_count
        )

        outcome = self.evaluate_threshold(threshold_value)

        return CheckResult(
            check=self._build_check_info(),
            outcome=outcome,
            threshold_value=threshold_value,
            diagnostic_metric_values=diagnostic_metric_values,
        )


class MultiColumnDistinctCountMetricImpl(AggregationMetricImpl):
    def __init__(
        self,
        contract_impl: ContractImpl,
        check_impl: MultiColumnDuplicateCheckImpl,
        column_names: list[str],
        data_source_impl: Optional[DataSourceImpl] = None,
        dataset_identifier: Optional[DatasetIdentifier] = None,
    ):
        self.column_names: list[str] = column_names
        super().__init__(
            contract_impl=contract_impl,
            metric_type="distinct_count",
            check_filter=check_impl.check_yaml.filter,
            data_source_impl=data_source_impl,
            dataset_identifier=dataset_identifier,
        )

    def _get_id_properties(self) -> dict[str, any]:
        id_properties: dict[str, str] = super()._get_id_properties()
        if isinstance(self.column_names, list):
            for index, column_name in enumerate(self.column_names):
                id_properties[f"column[{index}]"] = column_name
        return id_properties

    def sql_expression(self) -> SqlExpression:
        filter_expr: SqlExpression = SqlExpressionStr.optional(self.check_filter)
        if filter_expr:
            return COUNT(DISTINCT(CASE_WHEN(filter_expr, COMBINED_HASH(self.column_names))))
        else:
            return COUNT(DISTINCT(COMBINED_HASH(self.column_names)))

    def convert_db_value(self, value) -> int:
        # Note: expression SUM(CASE WHEN "id" IS NULL THEN 1 ELSE 0 END) gives NULL / None as a result if
        # there are no rows
        return int(value) if value is not None else 0


class MultiColumnDuplicateCountMetricImpl(DerivedMetricImpl):
    def __init__(
        self,
        metric_type: str,
        multi_column_distinct_count_metric_impl: MetricImpl,
        row_count_metric_impl: MetricImpl,
        check_filter: Optional[str],
    ):
        self.multi_column_distinct_count_metric_impl: MetricImpl = multi_column_distinct_count_metric_impl
        self.row_count_metric_impl: MetricImpl = row_count_metric_impl
        # Mind the ordering as the self._build_id() must come last
        super().__init__(
            contract_impl=multi_column_distinct_count_metric_impl.contract_impl,
            column_impl=multi_column_distinct_count_metric_impl.column_impl,
            metric_type=metric_type,
            check_filter=check_filter,
        )

    def get_metric_dependencies(self) -> list[MetricImpl]:
        return [self.multi_column_distinct_count_metric_impl, self.row_count_metric_impl]

    def compute_derived_value(self, measurement_values: MeasurementValues) -> Number:
        distinct_count: int = measurement_values.get_value(self.multi_column_distinct_count_metric_impl)
        row_count: int = measurement_values.get_value(self.row_count_metric_impl)
        return row_count - distinct_count



import hashlib
import logging
import os
import sys
from datetime import datetime
from logging import (
    CRITICAL,
    DEBUG,
    ERROR,
    INFO,
    WARNING,
    Formatter,
    LogRecord,
    StreamHandler,
)
from typing import Optional

from soda_core.common.logging_constants import Emoticons, ExtraKeys

verbose_mode: bool = False


def configure_logging(
    verbose: bool = False,
) -> None:
    """
    Used exposing the Soda log configurations.
    """
    global verbose_mode
    verbose_mode = verbose

    _prepare_masked_file()

    sys.stderr = sys.stdout
    for logger_to_mute in [
        "urllib3",
        "botocore",
        "pyathena",
        "faker",
        "snowflake",
        "matplotlib",
        "pyspark",
        "pyhive",
        "py4j",
        "segment",
    ]:
        logging.getLogger(logger_to_mute).setLevel(ERROR)

    soda_log_level: int = DEBUG if verbose else INFO

    logging.basicConfig(
        level=soda_log_level,
        force=True,  # Override any previously set handlers.
        # https://docs.python.org/3/library/logging.html#logrecord-attributes
        handlers=[SodaConsoleHandler()],
    )


_masked_values = set()


def _prepare_masked_file():
    global _masked_values
    file_with_masked_values = os.environ.get("SODA_MASKED_VALUES_FILE", None)
    if file_with_masked_values is not None:
        if not os.path.exists(file_with_masked_values):
            raise RuntimeError(f"Masked values file '{file_with_masked_values}' does not exist")
        sha_expected = os.environ.get("SODA_MASKED_VALUES_FILE_HASH", None)
        if not sha_expected:
            raise RuntimeError(f"Hash for the masked values file '{file_with_masked_values}' is not present")
        sha_hash = hashlib.sha256()
        with open(file_with_masked_values) as f:
            _masked_values.clear()
            read_lines = [l for l in f.readlines()]
            for l in read_lines:
                sha_hash.update(l.encode("utf-8"))
            sha_actual = sha_hash.hexdigest()
            if sha_actual != sha_expected:
                raise RuntimeError(
                    f"Hash mismatch for the masked values file '{file_with_masked_values}' (expected '{sha_expected}', got '{sha_actual}')"
                )
            _masked_values.update({l.strip() for l in read_lines})


def _mask_record(record: LogRecord):
    message = record.getMessage()
    if message:
        for masked in _masked_values:
            if (not masked) or masked not in message:
                continue
            message = message.replace(masked, "***")
    record.msg = message
    # since getMessage evaluates args, we need to clear them after the full message has been cleared
    record.args = ()
    if record.exc_info:
        exc_type, exc_value, exc_tb = record.exc_info
        masked_exc = _mask_exception(exc_value)
        record.exc_info = (exc_type, masked_exc, exc_tb)
    else:
        record.exc_info = None


def _mask_message(message: Optional[str]) -> Optional[str]:
    if not message or not _masked_values:
        return message
    for masked in _masked_values:
        if (not masked) or masked not in message:
            continue
        message = message.replace(masked, "***")
    return message


def _mask_exception(e: Optional[BaseException]) -> Optional[BaseException]:
    if e is None:
        return None

    # Mask the exception's args (usually contains the error message)
    if e.args:
        masked_args = []
        for arg in e.args:
            if isinstance(arg, str):
                masked_arg = _mask_message(arg)
                masked_args.append(masked_arg)
            else:
                masked_args.append(arg)
        e.args = tuple(masked_args)

    # Mask the traceback frames
    tb = e.__traceback__
    while tb is not None:
        frame = tb.tb_frame
        # Mask local variables in the frame that are strings
        if frame.f_locals:
            for var_name, var_value in frame.f_locals.items():
                if isinstance(var_value, str):
                    masked_value = _mask_message(var_value)
                    frame.f_locals[var_name] = masked_value
        tb = tb.tb_next

    # Recursively mask chained exceptions
    if e.__cause__ is not None:
        _mask_exception(e.__cause__)
    if e.__context__ is not None:
        _mask_exception(e.__context__)
    return e


def is_verbose() -> bool:
    return verbose_mode


class SodaConsoleHandler(StreamHandler):
    def __init__(self):
        super().__init__(sys.stdout)
        self.setFormatter(SodaConsoleFormatter())


class SodaConsoleFormatter(Formatter):
    def __init__(self):
        super().__init__()

    def format(self, record) -> str:
        parts: list[str] = [
            # self.format_timestamp(record),
            # self.format_level(record),
            self.format_message(record),
            self.format_location(record),
            self.format_doc(record),
            self.format_exception(record),
        ]
        return " | ".join(part for part in parts if part is not None)

    level_names: dict[int, str] = {
        CRITICAL: "CRI",
        ERROR: "ERR",
        WARNING: "WAR",
        INFO: "INF",
        DEBUG: "DEB",
    }

    def format_level(self, record: LogRecord) -> Optional[str]:
        return self.level_names.get(record.levelno, "UNKNOWN")

    def format_message(self, record: LogRecord) -> str:
        _mask_record(record)
        if record.levelno >= ERROR:
            return f"{Emoticons.POLICE_CAR_LIGHT} {record.getMessage()}"
        else:
            return record.getMessage()

    def format_timestamp(self, record: LogRecord) -> Optional[str]:
        timestamp: datetime = datetime.fromtimestamp(record.created)
        # Format the time part (without milliseconds)
        time_part = timestamp.strftime("%Y-%m-%d %H:%M:%S")
        # Add milliseconds with comma as separator
        milliseconds = timestamp.microsecond // 1000  # Convert microseconds to milliseconds
        return f"{time_part},{milliseconds:03d}"

    def format_location(self, record: LogRecord) -> Optional[str]:
        if hasattr(record, ExtraKeys.LOCATION):
            return str(record.location)
        else:
            return None

    def format_doc(self, record: LogRecord) -> Optional[str]:
        if hasattr(record, ExtraKeys.DOC):
            return record.doc
        else:
            return None

    def format_exception(self, record: LogRecord) -> Optional[str]:
        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = _mask_message(self.formatException(record.exc_info))
                return record.exc_text

        if hasattr(record, ExtraKeys.EXCEPTION):
            return record.exception
